from __future__ import annotations

import argparse
import contextlib
import os
import pathlib
import platform
import shlex
import shutil
import signal
import socket
import subprocess
import sys
import tempfile
import textwrap
import threading

import memray
from memray._errors import MemrayCommandError

from .live import LiveCommand
from .run import _get_free_port

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

TrackingMode = Literal["ACTIVATE", "DEACTIVATE", "FOR_DURATION"]


GDB_SCRIPT = pathlib.Path(__file__).parent / "_attach.gdb"
LLDB_SCRIPT = pathlib.Path(__file__).parent / "_attach.lldb"
RTLD_DEFAULT = memray._memray.RTLD_DEFAULT
RTLD_NOW = memray._memray.RTLD_NOW
PAYLOAD = """
import atexit
import time
import threading
import resource
import sys
from contextlib import suppress

import memray


class BareExceptionMessage(Exception):
    def __repr__(self):
        return self.args[0]


class RepeatingTimer(threading.Thread):
    def __init__(self, interval, function):
        self._interval = interval
        self._function = function
        self._canceled = threading.Event()
        super().__init__()

    def cancel(self):
        self._canceled.set()

    def run(self):
        while not self._canceled.wait(self._interval):
            if self._function():
                break


def deactivate_last_tracker():
    tracker = getattr(memray, "_last_tracker", None)
    if not tracker:
        return

    memray._last_tracker = None
    try:
        tracker.__exit__(None, None, None)
    finally:
        # Clean up resources associated with the Tracker ASAP,
        # even if an exception was raised.
        del tracker

    # Stop any waiting threads. This attribute may be unset if an old Memray
    # version attached 1st, setting last_tracker but not _attach_event_threads.
    # It could also be unset if we're racing another deactivate call.
    for thread in memray.__dict__.pop("_attach_event_threads", []):
        thread.cancel()


def activate_tracker():
    deactivate_last_tracker()
    tracker = {tracker_call}
    try:
        tracker.__enter__()
        memray._last_tracker = tracker
    finally:
        # Clean up resources associated with the Tracker ASAP,
        # even if an exception was raised.
        del tracker
    memray._attach_event_threads = []


def track_for_duration(duration=5):
    activate_tracker()

    def deactivate_because_timer_elapsed():
        print(
            "memray: Deactivating tracking:",
            duration,
            "seconds have elapsed",
            file=sys.stderr,
        )
        deactivate_last_tracker()

    thread = threading.Timer(duration, deactivate_because_timer_elapsed)
    thread.daemon = False
    thread.start()
    memray._attach_event_threads.append(thread)


if not hasattr(memray, "_last_tracker"):
    # This only needs to be registered the first time we attach.
    atexit.register(deactivate_last_tracker)

if {mode!r} == "ACTIVATE":
    activate_tracker()
elif {mode!r} == "DEACTIVATE":
    if not getattr(memray, "_last_tracker", None):
        raise BareExceptionMessage("no previous `memray attach` call detected")
    deactivate_last_tracker()
elif {mode!r} == "FOR_DURATION":
    track_for_duration({duration})
"""


def debugger_inject(debugger: str, pid: int, port: int, verbose: bool) -> str | None:
    """Execute a file in a running Python process using a debugger."""
    injecter = pathlib.Path(memray.__file__).parent / "_inject.abi3.so"
    assert injecter.exists()

    gdb_cmd = [
        "gdb",
        "-batch",
        "-p",
        str(pid),
        "-nx",
        "-nw",
        "-iex=set auto-solib-add off",
        f"-ex=set $rtld_now={RTLD_NOW}",
        f'-ex=set $libpath="{injecter}"',
        f"-ex=set $port={port}",
        f"-x={GDB_SCRIPT}",
    ]

    lldb_cmd = [
        "lldb",
        "--batch",
        "-p",
        str(pid),
        "--no-lldbinit",
        "-o",
        f'expr char $libpath[]="{injecter}"',
        "-o",
        f"expr int $port={port}",
        "-o",
        f"expr void* $rtld_default=(void*){RTLD_DEFAULT}",
        "-o",
        f"expr int $rtld_now={RTLD_NOW}",
        "--source",
        f"{LLDB_SCRIPT}",
    ]

    cmd = gdb_cmd if debugger == "gdb" else lldb_cmd
    if verbose:
        if sys.version_info >= (3, 8):
            print("Debugger command line:", shlex.join(cmd))
        else:
            print("Debugger command line:", cmd)

    try:
        output = subprocess.check_output(cmd, text=True, stderr=subprocess.STDOUT)
        returncode = 0
    except subprocess.CalledProcessError as exc:
        output = exc.output
        returncode = exc.returncode

    if cmd is lldb_cmd:
        # A bug in lldb sometimes means processes stay stopped after it exits.
        # Send a signal to wake the process up. Ignore any errors: the process
        # may have died, or may have never existed, or may be owned by another
        # user, etc. Processes that aren't stopped will ignore this signal, so
        # this should be harmless, though it is a huge hack.
        with contextlib.suppress(OSError):
            os.kill(pid, signal.SIGCONT)

    if verbose:
        print(f"debugger return code: {returncode}")
        print(f"debugger output:\n{output}")

    command_output_lines = (
        line for line in output.splitlines() if not line.startswith(f"({debugger})")
    )
    if returncode == 0 and any(' "SUCCESS"' in line for line in command_output_lines):
        return None

    # An error occurred. Give the best message we can. This is hacky; we don't
    # have a good option besides parsing output from the debugger session.
    if "--help" in output:
        return (
            "The debugger failed to parse our command line arguments.\n"
            "Run with --verbose to see the error message."
        )

    if "error: attach failed: " in output or "ptrace: " in output:
        # We failed to attach to the given pid. A few likely reasons...
        errmsg = "Failed to attach a debugger to the process.\n"
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            return errmsg + "The given process ID does not exist."
        except PermissionError:
            return errmsg + "The given process ID is owned by a different user."

        return errmsg + "You most likely do not have permission to trace the process."

    if "MEMRAY: Attached to process." not in output:
        return (
            f"Failed to execute our {debugger} script.\n"
            "Run with --verbose to debug the failure."
        )

    if "MEMRAY: Checking if process is Python 3.7+." in output:
        if "MEMRAY: Process is Python 3.7+." not in output:
            return "The process does not seem to be running Python 3.7 or newer."

    return "An unexpected error occurred. Run with --verbose to debug the failure."


def remote_exec_inject(pid: int, port: int, verbose: bool, tmpdir: str) -> str | None:
    """Execute a file in a running Python process using sys.remote_exec."""
    script = textwrap.dedent(
        f"""
        from contextlib import closing
        from queue import Queue
        from socket import create_connection
        from threading import Thread

        exc = None

        def thread_body():
            try:
                exec(sockfile.read(), globals())
            except Exception as exc:
                globals()["exc"] = exc

        with closing(create_connection((None, {port})).makefile("rw")) as sockfile:
            thread = Thread(target=thread_body)
            thread.start()
            thread.join()
            if exc is not None:
                sockfile.write(repr(exc))
        """
    )
    script_path = pathlib.Path(tmpdir) / "attach.py"
    script_path.write_text(script, encoding="utf-8")
    try:
        getattr(sys, "remote_exec")(pid, script_path)
    except Exception as exc:
        return f"Failed to execute script in remote process: {exc!r}"
    return None


def inject(method: str, pid: int, port: int, verbose: bool, tmpdir: str) -> str | None:
    """Execute a file in a running Python process."""
    if method == "sys.remote_exec":
        return remote_exec_inject(pid, port, verbose=verbose, tmpdir=tmpdir)
    return debugger_inject(method, pid, port, verbose=verbose)


def _sys_remote_exec_available(verbose: bool) -> bool:
    if not hasattr(sys, "remote_exec"):
        if verbose:
            print("sys.remote_exec is not available in this Python version")
        return False
    return True


def _gdb_available(verbose: bool) -> bool:
    if not shutil.which("gdb"):
        if verbose:
            print("No gdb executable found")
        return False
    return True


def _lldb_available(verbose: int) -> bool:
    # We need a version of lldb that supports `--batch`. This should be lldb
    # 3.5.2 or newer, but the version string format doesn't appear consistent
    # between macOS and Linux, so it's safer to just check the help output to
    # make sure that option is documented.
    try:
        help_str = subprocess.check_output(
            ["lldb", "--help"],
            text=True,
            stderr=subprocess.STDOUT,
        )
    except FileNotFoundError:
        if verbose:
            print("No lldb executable found")
        return False
    except subprocess.CalledProcessError as exc:
        if verbose:
            print(f"`lldb --version` failed: {exc.output}")
        return False

    if "--batch" not in help_str:
        if verbose:
            print("lldb does not support --batch, which we require")
        return False

    return True


def debugger_available(debugger: str, verbose: bool = False) -> bool:
    return {
        "sys.remote_exec": _sys_remote_exec_available,
        "gdb": _gdb_available,
        "lldb": _lldb_available,
    }[debugger](verbose=verbose)


def recvall(sock: socket.socket) -> str:
    return b"".join(iter(lambda: sock.recv(4096), b"")).decode("utf-8")


class ErrorReaderThread(threading.Thread):
    def __init__(self, sock: socket.socket) -> None:
        self._sock = sock
        super().__init__()

    def run(self) -> None:
        try:
            err = recvall(self._sock)
        except OSError as e:
            err = f"Unexpected exception: {e!r}"

        if not err:
            self.error = None
            return

        self.error = err
        os.kill(os.getpid(), signal.SIGINT)


class _DebuggerCommand:
    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--method",
            help="Method to use for injecting commands into the remote process",
            type=str,
            default="auto",
            choices=["auto", "sys.remote_exec", "gdb", "lldb"],
        )

        parser.add_argument(
            "-v",
            "--verbose",
            help="Print verbose debugging information.",
            action="store_true",
        )

        parser.add_argument(
            "pid",
            help="Process id to affect",
            type=int,
        )

    def resolve_debugger(self, method: str, *, verbose: bool = False) -> str:
        if method == "auto":
            # Prefer gdb on Linux but lldb on macOS
            if platform.system() == "Linux":
                debuggers = ("sys.remote_exec", "gdb", "lldb")
            else:
                debuggers = ("sys.remote_exec", "lldb", "gdb")

            for debugger in debuggers:
                if debugger_available(debugger, verbose=verbose):
                    return debugger
            raise MemrayCommandError(
                "Cannot find a supported lldb or gdb executable"
                " and sys.remote_exec is not available.",
                exit_code=1,
            )
        elif not debugger_available(method, verbose=verbose):
            raise MemrayCommandError(
                "sys.remote_exec requires Python 3.14 or newer."
                if method == "sys.remote_exec"
                else f"Cannot find a supported {method} executable.",
                exit_code=1,
            )
        return method

    def inject_control_channel(
        self, method: str, pid: int, *, verbose: bool = False
    ) -> socket.socket:
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        with contextlib.closing(server), tempfile.TemporaryDirectory() as tmpdir:
            server.bind(("localhost", 0))
            server.listen(1)
            sidechannel_port = server.getsockname()[1]

            errmsg = inject(
                method, pid, sidechannel_port, verbose=verbose, tmpdir=tmpdir
            )
            if errmsg:
                raise MemrayCommandError(errmsg, exit_code=1)

            server.settimeout(10)
            try:
                return server.accept()[0]
            except TimeoutError:
                raise MemrayCommandError(
                    f"Timed out waiting for connection from pid {pid}",
                    exit_code=1,
                )


class AttachCommand(_DebuggerCommand):
    """Begin tracking allocations in an already-started process"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-o",
            "--output",
            metavar="FILE",
            help=(
                "Capture allocations into the given file"
                " instead of starting a live tracking session"
            ),
        )
        parser.add_argument(
            "-f",
            "--force",
            help="If the output file already exists, overwrite it",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--aggregate",
            help="Write aggregated stats to the output file instead of all allocations",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--native",
            help="Track native (C/C++) stack frames as well",
            action="store_true",
            dest="native",
            default=False,
        )
        parser.add_argument(
            "--follow-fork",
            action="store_true",
            help="Record allocations in child processes forked from the tracked script",
            default=False,
        )
        parser.add_argument(
            "--trace-python-allocators",
            action="store_true",
            help="Record allocations made by the pymalloc allocator",
            default=False,
        )
        compression = parser.add_mutually_exclusive_group()
        compression.add_argument(
            "--compress-on-exit",
            help="Compress the resulting file using lz4 after tracking completes",
            default=True,
            action="store_true",
        )
        compression.add_argument(
            "--no-compress",
            help="Do not compress the resulting file using lz4",
            default=False,
            action="store_true",
        )

        parser.add_argument(
            "--duration", type=int, help="Duration to track for (in seconds)"
        )

        super().prepare_parser(parser)

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        verbose = args.verbose
        mode: TrackingMode = "ACTIVATE"
        duration = None

        if args.duration:
            mode = "FOR_DURATION"
            duration = args.duration

        args.method = self.resolve_debugger(args.method, verbose=verbose)

        destination: memray.Destination
        if args.output:
            live_port = None
            destination = memray.FileDestination(
                path=os.path.abspath(args.output),
                overwrite=args.force,
                compress_on_exit=not args.no_compress,
            )
        else:
            live_port = _get_free_port()
            destination = memray.SocketDestination(server_port=live_port)

        if args.aggregate and not args.output:
            parser.error("Can't use aggregated mode without an output file.")

        file_format = (
            "file_format=memray.FileFormat.AGGREGATED_ALLOCATIONS"
            if args.aggregate
            else ""
        )

        tracker_call = (
            f"memray.Tracker(destination=memray.{destination!r},"
            f" native_traces={args.native},"
            f" follow_fork={args.follow_fork},"
            f" trace_python_allocators={args.trace_python_allocators},"
            f"{file_format})"
        )

        client = self.inject_control_channel(args.method, args.pid, verbose=verbose)
        client.sendall(
            PAYLOAD.format(
                tracker_call=tracker_call,
                mode=mode,
                duration=duration,
            ).encode("utf-8")
        )
        client.shutdown(socket.SHUT_WR)

        if not live_port:
            err = recvall(client)
            if err:
                raise MemrayCommandError(
                    f"Failed to start tracking in remote process: {err}",
                    exit_code=1,
                )
            return

        # If an error prevents the tracked process from binding a server to
        # live_port, the TUI will hang forever trying to connect. Handle this
        # by spawning a background thread that watches for an error report over
        # the side channel and raises a SIGINT to interrupt the TUI if it sees
        # one. This can race, though: in some cases the TUI will also see an
        # error (if no header is sent over the socket), and the background
        # thread may raise a SIGINT that we see only after the live TUI has
        # already exited. If so we must ignore the extra KeyboardInterrupt.
        error_reader = ErrorReaderThread(client)
        error_reader.start()
        live = LiveCommand()

        with contextlib.suppress(KeyboardInterrupt):
            try:
                try:
                    live.start_live_interface(live_port)
                finally:
                    # Note: may get a spurious KeyboardInterrupt!
                    error_reader.join()
            except (Exception, KeyboardInterrupt):
                remote_err = error_reader.error
                if not remote_err:
                    raise  # Propagate the exception

                raise MemrayCommandError(
                    f"Failed to start tracking in remote process: {remote_err}",
                    exit_code=1,
                ) from None


class DetachCommand(_DebuggerCommand):
    """End the tracking started by a previous ``memray attach`` call"""

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        verbose = args.verbose
        mode: TrackingMode = "DEACTIVATE"
        args.method = self.resolve_debugger(args.method, verbose=verbose)
        client = self.inject_control_channel(args.method, args.pid, verbose=verbose)

        client.sendall(
            PAYLOAD.format(
                tracker_call=None,
                mode=mode,
                duration=None,
            ).encode("utf-8")
        )
        client.shutdown(socket.SHUT_WR)

        err = recvall(client)
        if err:
            raise MemrayCommandError(
                f"Failed to stop tracking in remote process: {err}",
                exit_code=1,
            )



import argparse
import shlex
import tempfile
from pathlib import Path
from textwrap import dedent
from textwrap import indent

from IPython.core.error import UsageError
from IPython.core.magic import Magics
from IPython.core.magic import cell_magic
from IPython.core.magic import magics_class
from IPython.display import IFrame
from IPython.display import display
from rich import print as pprint

from memray import FileReader
from memray.commands.common import warn_if_not_enough_symbols
from memray.reporters.flamegraph import FlameGraphReporter

TEMPLATE = """\
from memray import Tracker, FileFormat
with Tracker(
    "{dump_file!s}",
    native_traces={native_traces},
    trace_python_allocators={trace_python_allocators},
    follow_fork={follow_fork},
    file_format=FileFormat.{file_format},
) as tracker:
{code}
"""


def argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="%%memray_flamegraph")
    parser.add_argument(
        "--native",
        help="Track native (C/C++) stack frames as well",
        action="store_true",
        dest="native",
        default=False,
    )
    parser.add_argument(
        "--follow-fork",
        action="store_true",
        help="Record allocations in child processes forked from the tracked script",
        default=False,
    )
    parser.add_argument(
        "--trace-python-allocators",
        action="store_true",
        help="Record allocations made by the pymalloc allocator",
        default=False,
    )
    alloc_type_group = parser.add_mutually_exclusive_group()
    alloc_type_group.add_argument(
        "--leaks",
        help="Show memory leaks, instead of peak memory usage",
        action="store_true",
        dest="show_memory_leaks",
        default=False,
    )
    alloc_type_group.add_argument(
        "--temporary-allocation-threshold",
        metavar="N",
        help=dedent(
            """
            Report temporary allocations, as opposed to leaked allocations
            or high watermark allocations.  An allocation is considered
            temporary if at most N other allocations occur before it is
            deallocated.  With N=0, an allocation is temporary only if it
            is immediately deallocated before any other allocation occurs.
            """
        ),
        action="store",
        dest="temporary_allocation_threshold",
        type=int,
        default=-1,
    )
    alloc_type_group.add_argument(
        "--temporary-allocations",
        help="Equivalent to --temporary-allocation-threshold=1",
        action="store_const",
        dest="temporary_allocation_threshold",
        const=1,
    )

    parser.add_argument(
        "--temporal",
        help=(
            "Generate a dynamic flame graph that can analyze"
            " allocations in a user-selected time range."
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--split-threads",
        help="Do not merge allocations across threads",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--inverted",
        help=(
            "Invert the flame graph: "
            "use allocators as roots instead of thread entry points"
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--max-memory-records",
        help="Maximum number of memory records to display",
        type=int,
        default=None,
    )

    return parser


@magics_class
class FlamegraphMagics(Magics):
    @cell_magic  # type: ignore
    def memray_flamegraph(self, line: str, cell: str) -> None:
        """Memory profile the code in the cell and display a flame graph."""
        if self.shell is None:
            raise UsageError("Cannot profile code when not in a shell")

        try:
            options = argument_parser().parse_args(shlex.split(line))
        except SystemExit:
            # argparse wants to bail if the options aren't valid.
            # It already printed a message, just return control to IPython.
            return

        if options.temporal and options.temporary_allocation_threshold >= 0:
            raise UsageError(
                "Can't create a temporal flame graph of temporary allocations"
            )
        elif options.temporal or options.temporary_allocation_threshold >= 0:
            file_format = "ALL_ALLOCATIONS"
        else:
            file_format = "AGGREGATED_ALLOCATIONS"

        results_dir = Path("memray-results")
        results_dir.mkdir(exist_ok=True)

        tempdir = Path(tempfile.mkdtemp(dir=results_dir))
        dump_file = Path(tempdir) / "memray.dump"
        code = TEMPLATE.format(
            dump_file=dump_file,
            native_traces=options.native,
            trace_python_allocators=options.trace_python_allocators,
            follow_fork=options.follow_fork,
            code=indent(cell, " " * 4),
            file_format=file_format,
        )
        self.shell.run_cell(code)

        merge_threads = not options.split_threads

        reporter = None

        kwargs = {}
        if options.max_memory_records is not None:
            kwargs["max_memory_records"] = options.max_memory_records

        with FileReader(dump_file, report_progress=True, **kwargs) as reader:
            if reader.metadata.has_native_traces:
                warn_if_not_enough_symbols()

            if options.temporal:
                if options.show_memory_leaks:
                    temporal_snapshot = reader.get_temporal_allocation_records(
                        merge_threads=merge_threads
                    )
                    reporter = FlameGraphReporter.from_temporal_snapshot(
                        temporal_snapshot,
                        memory_records=tuple(reader.get_memory_snapshots()),
                        native_traces=reader.metadata.has_native_traces,
                        high_water_mark_by_snapshot=None,
                        inverted=options.inverted,
                    )
                else:
                    recs, hwms = reader.get_temporal_high_water_mark_allocation_records(
                        merge_threads=merge_threads
                    )
                    reporter = FlameGraphReporter.from_temporal_snapshot(
                        recs,
                        memory_records=tuple(reader.get_memory_snapshots()),
                        native_traces=reader.metadata.has_native_traces,
                        high_water_mark_by_snapshot=hwms,
                        inverted=options.inverted,
                    )
            else:
                if options.show_memory_leaks:
                    snapshot = reader.get_leaked_allocation_records(
                        merge_threads=merge_threads
                    )
                elif options.temporary_allocation_threshold >= 0:
                    snapshot = reader.get_temporary_allocation_records(
                        threshold=options.temporary_allocation_threshold,
                        merge_threads=merge_threads,
                    )
                else:
                    snapshot = reader.get_high_watermark_allocation_records(
                        merge_threads=merge_threads
                    )

                memory_records = tuple(reader.get_memory_snapshots())
                reporter = FlameGraphReporter.from_snapshot(
                    snapshot,
                    memory_records=memory_records,
                    native_traces=options.native,
                    inverted=options.inverted,
                )

        assert reporter is not None
        flamegraph_path = Path(tempdir) / "flamegraph.html"
        with open(flamegraph_path, "w") as f:
            reporter.render(
                outfile=f,
                metadata=reader.metadata,
                show_memory_leaks=options.show_memory_leaks,
                merge_threads=merge_threads,
                inverted=options.inverted,
            )
        dump_file.unlink()
        pprint(f"Results saved to [bold cyan]{flamegraph_path}")
        display(IFrame(flamegraph_path, width="100%", height="600"))


assert FlamegraphMagics.memray_flamegraph.__doc__ is not None
FlamegraphMagics.memray_flamegraph.__doc__ += "\n\n" + argument_parser().format_help()



import enum
from pathlib import Path
from types import CodeType
from types import FrameType
from types import TracebackType
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypedDict
from typing import Union
from typing import overload

from memray._destination import FileDestination as FileDestination
from memray._destination import SocketDestination as SocketDestination
from memray._metadata import Metadata
from memray._stats import Stats

from . import Destination

PythonStackElement = Tuple[str, str, int]
NativeStackElement = Tuple[str, str, int]
MemorySnapshot = NamedTuple(
    "MemorySnapshot", [("time", int), ("rss", int), ("heap", int)]
)

def set_log_level(level: int) -> None: ...

class AllocationRecord:
    @property
    def address(self) -> int: ...
    @property
    def allocator(self) -> int: ...
    @property
    def n_allocations(self) -> int: ...
    @property
    def size(self) -> int: ...
    @property
    def stack_id(self) -> int: ...
    @property
    def tid(self) -> int: ...
    @property
    def native_stack_id(self) -> int: ...
    @property
    def native_segment_generation(self) -> int: ...
    @property
    def thread_name(self) -> str: ...
    def hybrid_stack_trace(
        self,
        max_stacks: Optional[int] = None,
    ) -> List[Union[PythonStackElement, NativeStackElement]]: ...
    def native_stack_trace(
        self, max_stacks: Optional[int] = None
    ) -> List[NativeStackElement]: ...
    def stack_trace(
        self, max_stacks: Optional[int] = None
    ) -> List[PythonStackElement]: ...
    def __eq__(self, other: Any) -> Any: ...
    def __ge__(self, other: Any) -> Any: ...
    def __gt__(self, other: Any) -> Any: ...
    def __hash__(self) -> Any: ...
    def __le__(self, other: Any) -> Any: ...
    def __lt__(self, other: Any) -> Any: ...
    def __ne__(self, other: Any) -> Any: ...

class TrackedObjectRecord:
    @property
    def tid(self) -> int: ...
    @property
    def address(self) -> int: ...
    @property
    def is_created(self) -> bool: ...
    @property
    def frame_index(self) -> int: ...
    @property
    def native_frame_id(self) -> int: ...
    @property
    def native_segment_generation(self) -> int: ...
    def toPythonObject(self) -> Any: ...
    def hybrid_stack_trace(
        self,
        max_stacks: Optional[int] = None,
    ) -> List[Union[PythonStackElement, NativeStackElement]]: ...
    def native_stack_trace(
        self, max_stacks: Optional[int] = None
    ) -> List[NativeStackElement]: ...
    def stack_trace(
        self, max_stacks: Optional[int] = None
    ) -> List[PythonStackElement]: ...
    def __eq__(self, other: Any) -> Any: ...
    def __hash__(self) -> Any: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class Interval:
    def __init__(
        self,
        allocated_before_snapshot: int,
        deallocated_before_snapshot: int | None,
        n_allocations: int,
        n_bytes: int,
    ) -> None: ...
    def __eq__(self, other: Any) -> Any: ...
    allocated_before_snapshot: int
    deallocated_before_snapshot: int | None
    n_allocations: int
    n_bytes: int

class TemporalAllocationRecord:
    @property
    def allocator(self) -> int: ...
    @property
    def stack_id(self) -> int: ...
    @property
    def tid(self) -> int: ...
    @property
    def native_stack_id(self) -> int: ...
    @property
    def native_segment_generation(self) -> int: ...
    @property
    def thread_name(self) -> str: ...
    def hybrid_stack_trace(
        self,
        max_stacks: Optional[int] = None,
    ) -> List[Union[PythonStackElement, NativeStackElement]]: ...
    def native_stack_trace(
        self, max_stacks: Optional[int] = None
    ) -> List[NativeStackElement]: ...
    def stack_trace(
        self, max_stacks: Optional[int] = None
    ) -> List[PythonStackElement]: ...
    def __eq__(self, other: Any) -> Any: ...
    def __hash__(self) -> Any: ...
    intervals: List[Interval]

class AllocatorType(enum.IntEnum):
    PYMALLOC_FREE = 1
    PYMALLOC_MALLOC = 2
    PYMALLOC_CALLOC = 3
    PYMALLOC_REALLOC = 4
    FREE = 5
    MALLOC = 6
    REALLOC = 7
    CALLOC = 8
    POSIX_MEMALIGN = 9
    ALIGNED_ALLOC = 10
    MEMALIGN = 11
    VALLOC = 12
    PVALLOC = 13
    MMAP = 14
    MUNMAP = 15

class FileFormat(enum.IntEnum):
    ALL_ALLOCATIONS = 1
    AGGREGATED_ALLOCATIONS = 2

def start_thread_trace(frame: FrameType, event: str, arg: Any) -> None: ...

class FileReader:
    @property
    def metadata(self) -> Metadata: ...
    def __init__(
        self,
        file_name: Union[str, Path],
        *,
        report_progress: bool = False,
        max_memory_records: int = 10000,
    ) -> None: ...
    def get_allocation_records(self) -> Iterable[AllocationRecord]: ...
    def get_temporal_allocation_records(
        self,
        merge_threads: bool,
    ) -> Iterable[TemporalAllocationRecord]: ...
    def get_temporal_high_water_mark_allocation_records(
        self,
        merge_threads: bool,
    ) -> Tuple[List[TemporalAllocationRecord], List[int]]: ...
    def get_high_watermark_allocation_records(
        self,
        merge_threads: bool = ...,
    ) -> Iterable[AllocationRecord]: ...
    def get_leaked_allocation_records(
        self, merge_threads: bool = ...
    ) -> Iterable[AllocationRecord]: ...
    def get_temporary_allocation_records(
        self, merge_threads: bool = ..., threshold: int = ...
    ) -> Iterable[AllocationRecord]: ...
    def get_memory_snapshots(self) -> Iterable[MemorySnapshot]: ...
    def __enter__(self) -> Any: ...
    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> bool: ...
    @property
    def closed(self) -> bool: ...
    def close(self) -> None: ...
    def get_tracked_objects(
        self, filter_objs: Optional[Iterable[Any]] = None
    ) -> Iterable[TrackedObjectRecord]: ...

def compute_statistics(
    file_name: Union[str, Path],
    *,
    report_progress: bool = False,
    num_largest: int = 5,
) -> Stats: ...
def dump_all_records(file_name: Union[str, Path]) -> None: ...

class SocketReader:
    def __init__(self, port: int) -> None: ...
    def __enter__(self) -> "SocketReader": ...
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> Any: ...
    def get_current_snapshot(
        self, *, merge_threads: bool
    ) -> Iterator[AllocationRecord]: ...
    @property
    def command_line(self) -> Optional[str]: ...
    @property
    def is_active(self) -> bool: ...
    @property
    def pid(self) -> Optional[int]: ...
    @property
    def has_native_traces(self) -> bool: ...

class Tracker:
    @property
    def reader(self) -> FileReader: ...
    @overload
    def __init__(
        self,
        file_name: Union[Path, str],
        *,
        native_traces: bool = ...,
        memory_interval_ms: int = ...,
        follow_fork: bool = ...,
        trace_python_allocators: bool = ...,
        file_format: FileFormat = ...,
        reference_tracking: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        *,
        destination: Destination,
        native_traces: bool = ...,
        memory_interval_ms: int = ...,
        follow_fork: bool = ...,
        trace_python_allocators: bool = ...,
        file_format: FileFormat = ...,
        reference_tracking: bool = ...,
    ) -> None: ...
    def __enter__(self) -> Any: ...
    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> bool: ...

def greenlet_trace(event: str, args: Any) -> None: ...

class PymallocDomain(enum.IntEnum):
    PYMALLOC_RAW = 1
    PYMALLOC_MEM = 2
    PYMALLOC_OBJECT = 3

def size_fmt(num: int, suffix: str = "B") -> str: ...

class SymbolicSupport(enum.IntEnum):
    NONE = 1
    FUNCTION_NAME_ONLY = 2
    TOTAL = 3

def get_symbolic_support() -> SymbolicSupport: ...

RTLD_NOW: int
RTLD_DEFAULT: int

class HighWaterMarkAggregatorTestHarness:
    def add_allocation(
        self,
        tid: int,
        address: int,
        size: int,
        allocator: int,
        native_frame_id: int,
        frame_index: int,
        native_segment_generation: int,
    ) -> None: ...
    def capture_snapshot(self) -> None: ...
    def high_water_mark_bytes_by_snapshot(self) -> list[int]: ...
    def get_current_heap_size(self) -> int: ...
    def get_temporal_allocations(self) -> list[TemporalAllocationRecord]: ...
    def get_allocations(self) -> list[dict[str, int]]: ...

class AllocationLifetimeAggregatorTestHarness:
    def add_allocation(
        self,
        tid: int,
        address: int,
        size: int,
        allocator: int,
        native_frame_id: int,
        frame_index: int,
        native_segment_generation: int,
    ) -> None: ...
    def capture_snapshot(self) -> None: ...
    def get_allocations(self) -> list[TemporalAllocationRecord]: ...

class Segment(TypedDict):
    vaddr: int
    memsz: int

class Mapping(TypedDict):
    filename: str
    addr: int
    segments: list[Segment]

class RecordWriterTestHarness:
    def __init__(
        self,
        file_path: str,
        native_traces: bool = False,
        trace_python_allocators: bool = False,
        file_format: FileFormat = FileFormat.ALL_ALLOCATIONS,
        main_tid: int = 1,
        skipped_frames: int = 0,
        command_line: str = ...,
    ) -> None: ...
    def write_header(self, seek_to_start: bool) -> None: ...
    def write_memory_record(self, ms_since_epoch: int, rss: int) -> bool: ...
    def write_code_object(
        self,
        id: int,
        function_name: str,
        filename: str,
        linetable: bytes,
        firstlineno: int,
    ) -> bool: ...
    def write_unresolved_native_frame(self, ip: int, index: int) -> bool: ...
    def write_allocation_record(
        self,
        tid: int,
        address: int,
        size: int,
        allocator: int,
        native_frame_id: int = 0,
    ) -> bool: ...
    def write_frame_push(
        self,
        tid: int,
        code_object_id: int,
        instruction_offset: int,
        is_entry_frame: bool,
    ) -> bool: ...
    def write_frame_pop(self, tid: int, count: int) -> bool: ...
    def write_thread_record(self, tid: int, name: str) -> bool: ...
    def write_mappings(self, mappings: list[Mapping]) -> bool: ...
    def write_trailer(self) -> bool: ...
    @staticmethod
    def get_linetable(code_object: CodeType) -> bytes: ...
    @staticmethod
    def get_lasti(frame_object: FrameType) -> int: ...



import argparse

from ..reporters.table import TableReporter
from .common import HighWatermarkCommand


class TableCommand(HighWatermarkCommand):
    """Generate an HTML table with all records in the peak memory usage"""

    def __init__(self) -> None:
        super().__init__(
            reporter_factory=TableReporter.from_snapshot,
            reporter_name="table",
        )

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        super().prepare_parser(parser)
        parser.add_argument(
            "--no-web",
            help="Use local assets instead of fetching from CDN",
            action="store_true",
            default=False,
        )



import threading
from typing import Callable


class ThreadNameInterceptor:
    """Record the name of each threading.Thread for Memray's reports.

    The name can be set either before or after the thread is started, and from
    either the same thread or a different thread. Whenever an assignment to
    either `Thread._name` or `Thread._ident` is performed and the other has
    already been set, we call a callback with the thread's ident and name.
    """

    def __init__(self, attr: str, callback: Callable[[int, str], None]) -> None:
        self._attr = attr
        self._callback = callback

    def __set__(self, instance: threading.Thread, value: object) -> None:
        instance.__dict__[self._attr] = value
        ident = instance.__dict__.get("_ident")
        name = instance.__dict__.get("_name")
        if ident is not None and name is not None:
            self._callback(ident, name)



from typing import Any


def load_ipython_extension(ipython: Any) -> None:
    from .flamegraph import FlamegraphMagics

    ipython.register_magics(FlamegraphMagics)


__all__ = ["load_ipython_extension"]



import collections
import contextlib
import os
import pathlib
import platform
import sys

cimport cython

import threading
from datetime import datetime

from rich import print as pprint
from rich.progress import Progress
from rich.progress import SpinnerColumn

from posix.time cimport CLOCK_MONOTONIC
from posix.time cimport clock_gettime
from posix.time cimport timespec

from _memray cimport records
from _memray.algorithm cimport count
from _memray.hooks cimport Allocator
from _memray.hooks cimport isDeallocator
from _memray.logging cimport setLogThreshold
from _memray.native_resolver cimport unwindHere
from _memray.record_reader cimport RecordReader
from _memray.record_reader cimport RecordResult
from _memray.record_writer cimport PyCodeObject
from _memray.record_writer cimport PyFrameObject
from _memray.record_writer cimport RecordWriter
from _memray.record_writer cimport codeGetLinetable
from _memray.record_writer cimport createRecordWriter
from _memray.record_writer cimport frameGetLasti
from _memray.records cimport AggregatedAllocation
from _memray.records cimport Allocation as _Allocation
from _memray.records cimport FileFormat as _FileFormat
from _memray.records cimport MemoryRecord
from _memray.records cimport MemorySnapshot as _MemorySnapshot
from _memray.records cimport TrackedObject
from _memray.sink cimport FileSink
from _memray.sink cimport NullSink
from _memray.sink cimport Sink
from _memray.sink cimport SocketSink
from _memray.snapshot cimport NO_THREAD_INFO
from _memray.snapshot cimport AbstractAggregator
from _memray.snapshot cimport AggregatedCaptureReaggregator
from _memray.snapshot cimport AllocationLifetime
from _memray.snapshot cimport AllocationLifetimeAggregator
from _memray.snapshot cimport AllocationStatsAggregator
from _memray.snapshot cimport HighWatermark
from _memray.snapshot cimport HighWaterMarkAggregator
from _memray.snapshot cimport HighWatermarkFinder
from _memray.snapshot cimport HighWaterMarkLocationKey
from _memray.snapshot cimport Py_GetSnapshotAllocationRecords
from _memray.snapshot cimport Py_ListFromSnapshotAllocationRecords
from _memray.snapshot cimport SnapshotAllocationAggregator
from _memray.snapshot cimport TemporaryAllocationsAggregator
from _memray.socket_reader_thread cimport BackgroundSocketReader
from _memray.source cimport FileSource
from _memray.source cimport SocketSource
from _memray.tracking_api cimport RecursionGuard
from _memray.tracking_api cimport Tracker as NativeTracker
from _memray.tracking_api cimport install_trace_function
from _memray.tracking_api cimport set_up_pthread_fork_handlers
from cpython cimport Py_DECREF
from cpython cimport PyErr_CheckSignals
from cpython cimport PyObject
from libc.math cimport ceil
from libc.stdint cimport uint64_t
from libc.stdint cimport uintptr_t
from libcpp cimport bool
from libcpp.limits cimport numeric_limits
from libcpp.memory cimport make_shared
from libcpp.memory cimport make_unique
from libcpp.memory cimport shared_ptr
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string as cppstring
from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport move
from libcpp.utility cimport pair
from libcpp.vector cimport vector

from ._destination import Destination
from ._destination import FileDestination
from ._destination import SocketDestination
from ._metadata import Metadata
from ._stats import Stats
from ._thread_name_interceptor import ThreadNameInterceptor

# NOTE: We can't reinitialize tracking in a child process until the interpreter
#       has reinitialized its locks, so we do it in a Python fork handler.
#       But, Python fork handlers aren't guaranteed to run for every fork, and
#       the child process must never inherit an active tracker, so we must use
#       a pthread fork handler to disable tracking before forking.
set_up_pthread_fork_handlers()
os.register_at_fork(after_in_child=NativeTracker.childFork)


def set_log_level(int level):
    """Configure which log messages will be printed to stderr by memray.

    By default, only log records of severity `logging.WARNING` or higher will
    be printed, but you can adjust this threshold.

    Args:
        level (int): The lowest severity level that a log record can have and
            still be printed.
    """
    setLogThreshold(level)


cpdef enum AllocatorType:
    PYMALLOC_FREE = 1
    PYMALLOC_MALLOC = 2
    PYMALLOC_CALLOC = 3
    PYMALLOC_REALLOC = 4
    FREE = 5
    MALLOC = 6
    REALLOC = 7
    CALLOC = 8
    POSIX_MEMALIGN = 9
    ALIGNED_ALLOC = 10
    MEMALIGN = 11
    VALLOC = 12
    PVALLOC = 13
    MMAP = 14
    MUNMAP = 15

# Note: enumerator values are negative because they must be unique
#       from the enumerator values of the AllocatorType enum above
cpdef enum ObjectTrackingEvent:
    OBJECT_CREATED = -1
    OBJECT_DESTROYED = -2

cpdef enum PythonAllocatorType:
    PYTHON_ALLOCATOR_PYMALLOC = 1
    PYTHON_ALLOCATOR_PYMALLOC_DEBUG = 2
    PYTHON_ALLOCATOR_MALLOC = 3
    PYTHON_ALLOCATOR_OTHER = 4
    PYTHON_ALLOCATOR_MIMALLOC = 5
    PYTHON_ALLOCATOR_MIMALLOC_DEBUG = 6

cpdef enum FileFormat:
    ALL_ALLOCATIONS = _FileFormat.ALL_ALLOCATIONS
    AGGREGATED_ALLOCATIONS = _FileFormat.AGGREGATED_ALLOCATIONS


def size_fmt(num, suffix='B'):
    for unit in ['','k','M','G','T','P','E','Z']:
        if abs(num) < 1000.0:
            return f"{num:5.3f}{unit}{suffix}"
        num /= 1000.0
    return f"{num:.1f}Y{suffix}"

# Memray core

cdef stack_trace(
    RecordReader* reader,
    tid,
    allocator,
    python_stack_id,
    max_stacks=None,
):
    if allocator in (AllocatorType.FREE, AllocatorType.MUNMAP, ObjectTrackingEvent.OBJECT_DESTROYED):
        raise NotImplementedError("Stack traces for deallocations aren't captured.")

    assert reader != NULL, "Cannot get stack trace without reader."
    cdef ssize_t to_skip
    cdef ssize_t to_keep

    if max_stacks is not None:
        return reader.Py_GetStackFrame(python_stack_id, max_stacks)

    stack_trace = reader.Py_GetStackFrame(python_stack_id)
    if tid == reader.getMainThreadTid():
        to_skip = reader.getSkippedFramesOnMainThread()
        to_keep = max(len(stack_trace) - to_skip, 0)
        del stack_trace[to_keep:]
    return stack_trace


cdef native_stack_trace(
    RecordReader* reader,
    allocator,
    native_stack_id,
    generation,
    max_stacks=None,
):
    if allocator in (AllocatorType.FREE, AllocatorType.MUNMAP):
        raise NotImplementedError("Stack traces for deallocations aren't captured.")

    assert reader != NULL, "Cannot get stack trace without reader."
    if max_stacks is None:
        return reader.Py_GetNativeStackFrame(native_stack_id, generation)
    return reader.Py_GetNativeStackFrame(native_stack_id, generation, max_stacks)


cdef hybrid_stack_trace(
    RecordReader* reader,
    tid,
    allocator,
    python_stack_id,
    native_stack_id,
    generation,
    max_stacks=None,
):
    # This function merges a Python stack and a native stack into
    # a "hybrid" stack. It substitutes _PyFrame_EvalFrameDefault calls in
    # the native stack with the corresponding frame in the Python stack.
    # This sounds easy, but there are several tricky aspects:
    # 1. For the thread that called Tracker.__enter__, we want to hide
    #    frames (both Python and C) above the one that made that call.
    #    For other threads we want to keep all frames.
    # 2. If _PyFrame_EvalFrameDefault allocates memory before calling our
    #    profile function, we'll have too few Python frames to pair up
    #    every _PyFrame_EvalFrameDefault call. This happens in 3.11.
    # 3. Since Python 3.11, one _PyFrame_EvalFrameDefault call can evaluate
    #    many Python frames. If a frame's is_entry_frame flag is unset, it
    #    uses the same _PyFrame_EvalFrameDefault call as its caller.
    # 4. If the interpreter was stripped, we may not be able to recognize
    #    every (or even any) _PyFrame_EvalFrameDefault call, so we may
    #    have extra Python frames left after pairing.
    cdef vector[unsigned char] is_entry_frame
    native_stack = native_stack_trace(reader, allocator, native_stack_id, generation)
    python_stack = reader.Py_GetStackFrameAndEntryInfo(python_stack_id, &is_entry_frame)

    cdef ssize_t num_non_entry_frames = count(
        is_entry_frame.begin(), is_entry_frame.end(), 0
    )

    # Entry frames replace native frames; non-entry frames are inserted.
    hybrid_stack = [None] * (len(native_stack) + num_non_entry_frames)

    # Both stacks are from most recent to least, but we must pair things up
    # least recent to most to handle cases where _PyFrame_EvalFrameDefault
    # allocated memory before calling the profile function.
    native_stack.reverse()
    cdef ssize_t pidx = len(python_stack) - 1
    cdef ssize_t hidx = len(hybrid_stack) - 1

    cdef ssize_t to_skip = 0
    if tid == reader.getMainThreadTid():
        to_skip = reader.getSkippedFramesOnMainThread()
    cdef ssize_t first_kept_frame = pidx - to_skip

    for native_frame in native_stack:
        symbol = native_frame[0]
        if pidx >= 0 and "_PyEval_EvalFrameDefault" in symbol:
            while True:
                # If we're not keeping all frames and we've reached the
                # first one we want to keep, remove frames above it.
                if to_skip != 0 and pidx == first_kept_frame:
                    del hybrid_stack[hidx + 1:]

                assert hidx >= 0
                hybrid_stack[hidx] = python_stack[pidx]
                hidx -= 1
                pidx -= 1

                # Stop when we either run out of Python frames or reach the
                # entry frame being evaluated by the next eval loop.
                if pidx < 0 or is_entry_frame[pidx]:
                    break
        else:
            assert hidx >= 0
            hybrid_stack[hidx] = native_frame
            hidx -= 1

    if pidx >= 0:
        # We ran out of native frames without using up all of our Python
        # frames. We've seen this happen on stripped interpreters on Alpine
        # Linux in CI. Presumably this indicates that unwinding failed to
        # symbolify some of the calls to _PyEval_EvalFrameDefault.
        return python_stack
    assert hidx == -1

    return hybrid_stack[:max_stacks]


@cython.freelist(1024)
cdef class AllocationRecord:
    cdef object _tuple
    cdef dict _stack_trace_cache
    cdef shared_ptr[RecordReader] _reader

    def __init__(self, record):
        self._tuple = record
        self._stack_trace_cache = {}

    def __eq__(self, other):
        cdef AllocationRecord _other
        if isinstance(other, AllocationRecord):
            _other = other
            return self._tuple == _other._tuple
        return NotImplemented

    def __hash__(self):
        return hash(self._tuple)

    @property
    def tid(self):
        return self._tuple[0]

    @property
    def address(self):
        return self._tuple[1]

    @property
    def size(self):
        return self._tuple[2]

    @property
    def allocator(self):
        return self._tuple[3]

    @property
    def stack_id(self):
        return self._tuple[4]

    @property
    def n_allocations(self):
        return self._tuple[5]

    @property
    def native_stack_id(self):
        return self._tuple[6]

    @property
    def native_segment_generation(self):
        return self._tuple[7]

    @property
    def thread_name(self):
        if self.tid == -1:
            return "merged thread"
        assert self._reader.get() != NULL, "Cannot get thread name without reader."
        return self._reader.get().getThreadName(self.tid)

    def stack_trace(self, max_stacks=None):
        cache_key = ("python", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = stack_trace(
                self._reader.get(),
                self.tid,
                self.allocator,
                self.stack_id,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def native_stack_trace(self, max_stacks=None):
        cache_key = ("native", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = native_stack_trace(
                self._reader.get(),
                self.allocator,
                self.native_stack_id,
                self.native_segment_generation,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def hybrid_stack_trace(self, max_stacks=None):
        cache_key = ("hybrid", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = hybrid_stack_trace(
                self._reader.get(),
                self.tid,
                self.allocator,
                self.stack_id,
                self.native_stack_id,
                self.native_segment_generation,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def __repr__(self):
        return (f"AllocationRecord<tid={hex(self.tid)}, address={hex(self.address)}, "
                f"size={'N/A' if not self.size else size_fmt(self.size)}, allocator={self.allocator!r}, "
                f"allocations={self.n_allocations}>")


@cython.freelist(1024)
cdef class TrackedObjectRecord:
    cdef object _tuple
    cdef dict _stack_trace_cache
    cdef shared_ptr[RecordReader] _reader

    def __init__(self, record):
        self._tuple = record
        self._stack_trace_cache = {}

    def __eq__(self, other):
        cdef TrackedObjectRecord _other
        if isinstance(other, TrackedObjectRecord):
            _other = other
            return self._tuple == _other._tuple
        return NotImplemented

    def __hash__(self):
        return hash(self._tuple)

    @property
    def tid(self):
        return self._tuple[0]

    @property
    def address(self):
        return self._tuple[1]

    @property
    def is_created(self):
        return self._tuple[2]

    @property
    def stack_id(self):
        return self._tuple[3]

    @property
    def native_stack_id(self):
        return self._tuple[4]

    @property
    def native_segment_generation(self):
        return self._tuple[5]

    @property
    def thread_name(self):
        if self.tid == -1:
            return "merged thread"
        assert self._reader.get() != NULL, "Cannot get thread name without reader."
        return self._reader.get().getThreadName(self.tid)

    def stack_trace(self, max_stacks=None):
        cache_key = ("python", max_stacks)
        event = (
                ObjectTrackingEvent.OBJECT_CREATED if self.is_created
                else ObjectTrackingEvent.OBJECT_DESTROYED
        )
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = stack_trace(
                self._reader.get(),
                self.tid,
                event,
                self.stack_id,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def native_stack_trace(self, max_stacks=None):
        cache_key = ("native", max_stacks)
        event = (
                ObjectTrackingEvent.OBJECT_CREATED if self.is_created
                else ObjectTrackingEvent.OBJECT_DESTROYED
        )
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = native_stack_trace(
                self._reader.get(),
                event,
                self.native_stack_id,
                self.native_segment_generation,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def hybrid_stack_trace(self, max_stacks=None):
        cache_key = ("hybrid", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = hybrid_stack_trace(
                self._reader.get(),
                self.tid,
                None,  # No allocator for objects
                self.stack_id,
                self.native_stack_id,
                self.native_segment_generation,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def __repr__(self):
        return (f"TrackedObjectRecord<tid={hex(self.tid)}, address={hex(self.address)}, "
                f"is_created={self.is_created}, stack_id={self.stack_id}>")


@cython.freelist(1024)
cdef class Interval:
    cdef public size_t allocated_before_snapshot
    cdef public object deallocated_before_snapshot
    cdef public size_t n_allocations
    cdef public size_t n_bytes

    def __cinit__(
        self,
        size_t allocated_before_snapshot,
        object deallocated_before_snapshot,
        size_t n_allocations,
        size_t n_bytes,
    ):
        self.allocated_before_snapshot = allocated_before_snapshot
        self.deallocated_before_snapshot = deallocated_before_snapshot
        self.n_allocations = n_allocations
        self.n_bytes = n_bytes

    def __eq__(self, other):
        if type(other) is not Interval:
            return NotImplemented
        return (
            self.allocated_before_snapshot,
            self.deallocated_before_snapshot,
            self.n_allocations,
            self.n_bytes,
        ) == (
            other.allocated_before_snapshot,
            other.deallocated_before_snapshot,
            other.n_allocations,
            other.n_bytes,
        )

    def __repr__(self):
        return (
            f"Interval(allocated_before_snapshot={self.allocated_before_snapshot},"
            f" deallocated_before_snapshot={self.deallocated_before_snapshot},"
            f" n_allocations={self.n_allocations},"
            f" n_bytes={self.n_bytes})"
        )


@cython.freelist(1024)
cdef class TemporalAllocationRecord:
    cdef object _tuple
    cdef dict _stack_trace_cache
    cdef shared_ptr[RecordReader] _reader
    cdef public object intervals

    def __cinit__(self, record):
        self._tuple = record
        self._stack_trace_cache = {}
        self.intervals = []

    def __eq__(self, other):
        if type(other) != TemporalAllocationRecord:
            return NotImplemented
        cdef TemporalAllocationRecord o = other
        return self._tuple == o._tuple and self._intervals == o._intervals

    def __hash__(self):
        return hash(self._tuple)

    @property
    def tid(self):
        return self._tuple[0]

    @property
    def allocator(self):
        return self._tuple[1]

    @property
    def stack_id(self):
        return self._tuple[2]

    @property
    def native_stack_id(self):
        return self._tuple[3]

    @property
    def native_segment_generation(self):
        return self._tuple[4]

    @property
    def thread_name(self):
        assert self._reader.get() != NULL, "Cannot get thread name without reader."
        return self._reader.get().getThreadName(self.tid)

    def stack_trace(self, max_stacks=None):
        cache_key = ("python", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = stack_trace(
                self._reader.get(),
                self.tid,
                self.allocator,
                self.stack_id,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def native_stack_trace(self, max_stacks=None):
        cache_key = ("native", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = native_stack_trace(
                self._reader.get(),
                self.allocator,
                self.native_stack_id,
                self.native_segment_generation,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]

    def hybrid_stack_trace(self, max_stacks=None):
        cache_key = ("hybrid", max_stacks)
        if cache_key not in self._stack_trace_cache:
            self._stack_trace_cache[cache_key] = hybrid_stack_trace(
                self._reader.get(),
                self.tid,
                self.allocator,
                self.stack_id,
                self.native_stack_id,
                self.native_segment_generation,
                max_stacks,
            )
        return self._stack_trace_cache[cache_key]


cdef create_temporal_allocation_record(
    const HighWaterMarkLocationKey& key,
    shared_ptr[RecordReader] reader,
):
    cdef object elem = (
        key.thread_id,
        <int>key.allocator,
        key.python_frame_id,
        key.native_frame_id,
        key.native_segment_generation,
    )
    cdef TemporalAllocationRecord alloc = TemporalAllocationRecord(elem)
    alloc._reader = reader
    return alloc


cdef class TemporalAllocationGenerator:
    cdef vector[AllocationLifetime] lifetimes
    cdef shared_ptr[RecordReader] reader

    cdef TemporalAllocationRecord curr_record
    cdef HighWaterMarkLocationKey last_key
    cdef size_t idx

    cdef setup(
        self,
        vector[AllocationLifetime]&& lifetimes,
        shared_ptr[RecordReader] reader,
    ):
        self.lifetimes = move(lifetimes)
        self.reader = move(reader)

    def __iter__(self):
        return self

    def __next__(self):
        to_return = None
        cdef AllocationLifetime lifetime
        while self.idx < self.lifetimes.size():
            lifetime = self.lifetimes[self.idx]
            self.idx += 1

            if lifetime.key != self.last_key:
                if self.curr_record is not None:
                    to_return = self.curr_record
                self.last_key = lifetime.key
                self.curr_record = create_temporal_allocation_record(
                    lifetime.key, self.reader
                )

            if lifetime.deallocatedBeforeSnapshot == <size_t>(-1):
                deallocated_before_snapshot = None
            else:
                deallocated_before_snapshot = lifetime.deallocatedBeforeSnapshot

            self.curr_record.intervals.append(
                Interval(
                    lifetime.allocatedBeforeSnapshot,
                    deallocated_before_snapshot,
                    lifetime.n_allocations,
                    lifetime.n_bytes,
                )
            )

            if to_return is not None:
                return to_return

        if self.curr_record is not None:
            to_return = self.curr_record
            self.curr_record = None
            return to_return
        raise StopIteration


MemorySnapshot = collections.namedtuple("MemorySnapshot", "time rss heap")

tracker_creation_lock = threading.Lock()


cdef class Tracker:
    """Context manager for tracking memory allocations in a Python script.

    You can track memory allocations in a Python process by using a Tracker as
    a context manager::

        with memray.Tracker("some_output_file.bin"):

    Any code inside of the ``with`` block will have its allocations tracked.
    Any allocations made by other threads will also be tracked for the duration
    of the ``with`` block. Because of the way tracking works, there can only be
    one tracker active in the entire program at a time. Attempting to activate
    a tracker while one is already active will raise an exception, as will
    attempting to activate the same tracker more than once. If you want to
    re-enable tracking after the ``with`` block ends, you will need to create
    a fresh `Tracker` instance.

    Args:
        file_name (str or pathlib.Path): The name of the file to write the
            captured allocations into. This is the only argument that can be
            passed positionally. If not provided, the *destination* keyword
            argument must be provided.
        destination (FileDestination or SocketDestination): The destination to
            write captured allocations to. If provided, the *file_name*
            argument must not be provided.
        native_traces (bool): Whether or not to capture native stack frames, in
            addition to Python stack frames (see :ref:`Native Tracking`).
            Defaults to False.
        trace_python_allocators (bool): Whether or not to trace Python allocators
            as independent allocations. (see :ref:`Python allocators`).
            Defaults to False.
        follow_fork (bool): Whether or not to continue tracking in a subprocess
            that is forked from the tracked process (see :ref:`Tracking across
            Forks`). Defaults to False.
        memory_interval_ms (int): How many milliseconds to wait between sending
            periodic resident set size updates. By default, every 10
            milliseconds a record is written that contains the current
            timestamp and the total number of bytes of virtual memory allocated
            by the process. These records are used to create the graph of
            memory usage over time that appears at the top of the flame graph,
            for instance. This parameter lets you adjust the frequency between
            updates, though you shouldn't need to change it.
        file_format (FileFormat): The format that should be used when writing
            to the capture file. See the `FileFormat` documentation for a list
            of supported file formats and their limitations.
    """
    cdef bool _native_traces
    cdef bool _track_object_lifetimes
    cdef unsigned int _memory_interval_ms
    cdef bool _follow_fork
    cdef bool _trace_python_allocators
    cdef object _previous_profile_func
    cdef object _previous_thread_profile_func
    cdef unique_ptr[RecordWriter] _writer
    cdef object _surviving_objects

    cdef unique_ptr[Sink] _make_writer(self, destination) except*:
        # Creating a Sink can raise Python exceptions (if is interrupted by signal
        # handlers). If this happens, this method will propagate the appropriate exception.
        if isinstance(destination, FileDestination):
            is_dev_null = False
            with contextlib.suppress(OSError):
                if pathlib.Path("/dev/null").samefile(destination.path):
                    is_dev_null = True

            if is_dev_null:
                return unique_ptr[Sink](new NullSink())
            return unique_ptr[Sink](new FileSink(os.fsencode(destination.path),
                                                 destination.overwrite,
                                                 destination.compress_on_exit))

        elif isinstance(destination, SocketDestination):
            return unique_ptr[Sink](new SocketSink(destination.address, destination.server_port))
        else:
            raise TypeError("destination must be a SocketDestination or FileDestination")

    def __cinit__(self, object file_name=None, *, object destination=None,
                  bool native_traces=False, unsigned int memory_interval_ms = 10,
                  bool follow_fork=False, bool trace_python_allocators=False,
                  track_object_lifetimes=False, FileFormat file_format=FileFormat.ALL_ALLOCATIONS):
        if (file_name, destination).count(None) != 1:
            raise TypeError("Exactly one of 'file_name' or 'destination' argument must be specified")

        # Check Python version if reference tracking is enabled
        if track_object_lifetimes and sys.version_info < (3, 13, 3):
            raise RuntimeError(
                "Python object reference tracking requires Python 3.13.3 or later. "
                f"Current version: {'.'.join(map(str, sys.version_info[:3]))}"
            )

        cdef cppstring command_line = " ".join(sys.argv)
        self._native_traces = native_traces
        self._track_object_lifetimes = track_object_lifetimes
        self._memory_interval_ms = memory_interval_ms
        self._follow_fork = follow_fork
        self._trace_python_allocators = trace_python_allocators

        if file_name is not None:
            destination = FileDestination(path=file_name)

        if not isinstance(destination, FileDestination):
            if follow_fork:
                raise RuntimeError("follow_fork requires an output file")

            if file_format != FileFormat.ALL_ALLOCATIONS:
                raise RuntimeError("AGGREGATED_ALLOCATIONS requires an output file")

        self._writer = move(
            createRecordWriter(
                move(self._make_writer(destination)),
                command_line,
                native_traces,
                file_format,
                trace_python_allocators,
                track_object_lifetimes,
            )
        )

    @cython.profile(False)
    def __enter__(self):
        cdef unique_ptr[RecordWriter] writer
        with tracker_creation_lock:
            if NativeTracker.getTracker() != NULL:
                raise RuntimeError("No more than one Tracker instance can be active at the same time")

            if self._writer == NULL:
                raise RuntimeError("Attempting to use stale output handle")
            writer = move(self._writer)

            for attr in ("_name", "_ident"):
                assert not hasattr(threading.Thread, attr)
                setattr(
                    threading.Thread,
                    attr,
                    ThreadNameInterceptor(attr, NativeTracker.registerThreadNameById),
                )

            orig_set_os_name = getattr(threading.Thread, "_set_os_name", None)
            if orig_set_os_name is not None:
                def set_os_name_wrapper(self):
                    cdef unique_ptr[RecursionGuard] guard = make_unique[RecursionGuard]()
                    orig_set_os_name(self)

                setattr(threading.Thread, "_set_os_name", set_os_name_wrapper)

            self._previous_profile_func = sys.getprofile()
            self._previous_thread_profile_func = threading._profile_hook
            threading.setprofile(start_thread_trace)

            if "greenlet" in sys.modules:
                NativeTracker.beginTrackingGreenlets()

            self._surviving_objects = []

            NativeTracker.createTracker(
                move(writer),
                self._native_traces,
                self._memory_interval_ms,
                self._follow_fork,
                self._trace_python_allocators,
                self._track_object_lifetimes,
            )
            return self

    @cython.profile(False)
    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._populate_suriving_objects()
        with tracker_creation_lock:
            NativeTracker.destroyTracker()
            sys.setprofile(self._previous_profile_func)
            threading.setprofile(self._previous_thread_profile_func)

            for attr in ("_name", "_ident"):
                delattr(threading.Thread, attr)

    cdef void _populate_suriving_objects(self):
        assert NativeTracker.getTracker() != NULL
        cdef unordered_set[PyObject*] objects = NativeTracker.getTracker().getSurvivingObjects()
        for obj in objects:
            self._surviving_objects.append(<object>obj)
            Py_DECREF(<object>obj)

    def get_surviving_objects(self):
        """Get a list of objects that were alive at the end of the tracking period.

        Returns:
            list: A list of objects that were alive at the end of the tracking period.
        """
        if not self._track_object_lifetimes:
            raise RuntimeError(
                "track_object_lifetimes=True was not provided at Tracker construction"
            )
        return tuple(self._surviving_objects)


def start_thread_trace(frame, event, arg):
    if event in {"call", "c_call"}:
        install_trace_function()
    return start_thread_trace


def greenlet_trace_function(event, args):
    if event in {"switch", "throw"}:
        NativeTracker.handleGreenletSwitch(args[0], args[1])


def print_greenlet_warning():
    pprint(
        ":warning: [bold red]Memray support for Greenlet is experimental[/] :warning:\n"
        "[yellow]Please report any issues at https://github.com/bloomberg/memray/issues[/]\n"
        "\n",
        file=sys.stderr,
    )


cdef millis_to_dt(millis):
    return datetime.fromtimestamp(millis // 1000).replace(
        microsecond=millis % 1000 * 1000).astimezone()


cdef _create_metadata(header, peak_memory):
    stats = header["stats"]
    allocator_id_to_name = {
        PythonAllocatorType.PYTHON_ALLOCATOR_PYMALLOC: "pymalloc",
        PythonAllocatorType.PYTHON_ALLOCATOR_PYMALLOC_DEBUG: "pymalloc debug",
        PythonAllocatorType.PYTHON_ALLOCATOR_MIMALLOC: "mimalloc",
        PythonAllocatorType.PYTHON_ALLOCATOR_MIMALLOC_DEBUG: "mimalloc debug",
        PythonAllocatorType.PYTHON_ALLOCATOR_MALLOC: "malloc",
        PythonAllocatorType.PYTHON_ALLOCATOR_OTHER: "unknown",
    }
    return Metadata(
        start_time=millis_to_dt(stats["start_time"]),
        end_time=millis_to_dt(stats["end_time"]),
        total_allocations=stats["n_allocations"],
        total_frames=stats["n_frames"],
        peak_memory=peak_memory,
        command_line=header["command_line"],
        pid=header["pid"],
        main_thread_id=header["main_tid"],
        python_allocator=allocator_id_to_name[header["python_allocator"]],
        has_native_traces=header["native_traces"],
        trace_python_allocators=header["trace_python_allocators"],
        file_format=FileFormat(header["file_format"]),
    )


cdef class ProgressIndicator:
    cdef bool _report_progress
    cdef object _indicator
    cdef object _context_manager
    cdef object _task
    cdef object _task_description
    cdef object _total
    cdef size_t _cumulative_num_processed
    cdef size_t _update_interval
    cdef size_t _ns_between_refreshes
    cdef timespec _next_refresh

    def __init__(self,
        str task_description,
        object total,
        bool report_progress=True,
        size_t refresh_per_second=10,
    ):
        self._report_progress = report_progress
        self._total = total
        self._cumulative_num_processed = 0
        # Only check the elapsed time every N records
        self._update_interval = 100_000
        self._ns_between_refreshes = 1_000_000_000 // refresh_per_second
        self._next_refresh.tv_sec = 0
        self._next_refresh.tv_nsec = 0
        self._task_description = task_description
        self._task = None
        self._context_manager = None
        if report_progress:
            self._indicator = Progress(
                SpinnerColumn(),
                *Progress.get_default_columns(),
                auto_refresh=False,
                transient=True,
            )

    def __enter__(self):
        if not self._report_progress:
            return self
        self._context_manager = self._indicator.__enter__()
        self._task = self._context_manager.add_task(
            f"[cyan]{self._task_description}...",
            total=self._total
        )
        return self

    def __exit__(self, type, value, traceback):
        if not self._report_progress:
            return
        self._context_manager.update(self._task, completed=self._cumulative_num_processed)
        return self._context_manager.__exit__(type, value, traceback)

    cdef bool _time_for_refresh(self):
        cdef timespec now
        cdef int rc = clock_gettime(CLOCK_MONOTONIC, &now)

        if 0 != rc:
            return True

        if now.tv_sec > self._next_refresh.tv_sec or (
            now.tv_sec == self._next_refresh.tv_sec
            and now.tv_nsec > self._next_refresh.tv_nsec
        ):
            self._next_refresh = now
            self._next_refresh.tv_nsec += self._ns_between_refreshes
            while self._next_refresh.tv_nsec > 1_000_000_000:
                self._next_refresh.tv_nsec -= 1_000_000_000
                self._next_refresh.tv_sec += 1
            return True

        return False

    cdef update(self, size_t n_processed):
        self._cumulative_num_processed += n_processed
        if not self._report_progress:
            return
        if self._cumulative_num_processed % self._update_interval == 0:
            if self._time_for_refresh():
                assert(self._context_manager is not None)
                self._context_manager.update(
                    self._task, completed=self._cumulative_num_processed
                )
                self._context_manager.refresh()

    @property
    def num_processed(self):
        return self._cumulative_num_processed


cdef class FileReader:
    cdef cppstring _path

    cdef object _file
    cdef vector[_MemorySnapshot] _memory_snapshots
    cdef HighWatermark _high_watermark
    cdef object _header
    cdef bool _report_progress
    cdef size_t _memory_snapshot_stride

    def __cinit__(self, object file_name, *, bool report_progress=False, int max_memory_records=10000):
        try:
            self._file = open(file_name)
        except OSError as exc:
            raise OSError(f"Could not open file {file_name}: {exc.strerror}") from None

        if platform.system() == "Linux":
            self._path = "/proc/self/fd/" + str(self._file.fileno())
        else:
            self._path = str(file_name)
        self._report_progress = report_progress

        # Initial pass to populate _header, _high_watermark, and _memory_snapshots.
        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path)),
            False
        )
        cdef RecordReader* reader = reader_sp.get()

        self._header = reader.getHeader()
        stats = self._header["stats"]

        n_memory_snapshots_approx = 2048
        if 0 < stats["start_time"] < stats["end_time"]:
            n_memory_snapshots_approx = (stats["end_time"] - stats["start_time"]) / 10

        if n_memory_snapshots_approx > max_memory_records:
            n_memory_snapshots_approx = max_memory_records
        self._memory_snapshots.reserve(n_memory_snapshots_approx)

        cdef object total = stats['n_allocations'] or None
        cdef HighWatermarkFinder finder

        cdef ProgressIndicator progress_indicator = ProgressIndicator(
            "Calculating high watermark",
            total=total,
            report_progress=self._report_progress
        )
        self._memory_snapshot_stride = 0
        cdef MemoryRecord memory_record
        with progress_indicator:
            while True:
                PyErr_CheckSignals()
                ret = reader.nextRecord()
                if ret == RecordResult.RecordResultAllocationRecord:
                    finder.processAllocation(reader.getLatestAllocation())
                    progress_indicator.update(1)
                elif ret == RecordResult.RecordResultAggregatedAllocationRecord:
                    finder.processAllocation(
                        reader.getLatestAggregatedAllocation().contributionToHighWaterMark()
                    )
                    progress_indicator.update(1)
                elif ret == RecordResult.RecordResultMemoryRecord:
                    memory_record = reader.getLatestMemoryRecord()
                    self._memory_snapshots.push_back(
                        _MemorySnapshot(
                            memory_record.ms_since_epoch,
                            memory_record.rss,
                            finder.getCurrentWatermark(),
                        )
                    )
                elif ret == RecordResult.RecordResultMemorySnapshot:
                    self._memory_snapshots.push_back(reader.getLatestMemorySnapshot())
                else:
                    break

        if len(self._memory_snapshots) > max_memory_records:
            self._memory_snapshot_stride = int(ceil(<double>len(self._memory_snapshots) / max_memory_records))
            self._memory_snapshots = self._memory_snapshots[::self._memory_snapshot_stride]
        self._high_watermark = finder.getHighWatermark()
        stats["n_allocations"] = progress_indicator.num_processed

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self._file is not None:
            file = self._file
            self._file = None
            file.close()

    cdef void _ensure_not_closed(self) except *:
        if self._file is None:
            raise ValueError("Operation on a closed FileReader")

    @property
    def closed(self):
        return self._file is None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def _reaggregate_allocations(
        self, size_t records_to_process, bool merge_threads, bool leaks
    ):
        """Aggregate records from an AGGREGATED_ALLOCATIONS capture file.

        An extra aggregation pass is still needed to account for merge_threads,
        as well as (for now at least) different location key formats.
        """
        cdef AggregatedCaptureReaggregator aggregator
        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path))
        )
        cdef RecordReader* reader = reader_sp.get()

        cdef ProgressIndicator progress_indicator = ProgressIndicator(
            "Processing allocation records",
            total=records_to_process,
            report_progress=self._report_progress
        )

        cdef AggregatedAllocation record
        with progress_indicator:
            while True:
                PyErr_CheckSignals()
                ret = reader.nextRecord()
                if ret == RecordResult.RecordResultAggregatedAllocationRecord:
                    record = reader.getLatestAggregatedAllocation()
                    if leaks:
                        aggregator.addAllocation(record.contributionToLeaks())
                    else:
                        aggregator.addAllocation(record.contributionToHighWaterMark())
                    records_to_process -= 1
                    progress_indicator.update(1)
                elif ret == RecordResult.RecordResultMemorySnapshot:
                    pass
                else:
                    assert ret != RecordResult.RecordResultMemoryRecord
                    assert ret != RecordResult.RecordResultAllocationRecord
                    break

        for elem in Py_ListFromSnapshotAllocationRecords(
            aggregator.getSnapshotAllocations(merge_threads)
        ):
            alloc = AllocationRecord(elem)
            (<AllocationRecord> alloc)._reader = reader_sp
            yield alloc

        reader.close()

    def _aggregate_allocations(self, size_t records_to_process, bool merge_threads,
                               size_t temporary_buffer_size=0):
        cdef unique_ptr[AbstractAggregator] the_aggregator
        if temporary_buffer_size:
            the_aggregator.reset(
                new TemporaryAllocationsAggregator(temporary_buffer_size)
            )
        else:
            the_aggregator.reset(new SnapshotAllocationAggregator())
        cdef AbstractAggregator* aggregator = the_aggregator.get()

        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path))
        )
        cdef RecordReader* reader = reader_sp.get()

        cdef ProgressIndicator progress_indicator = ProgressIndicator(
            "Processing allocation records",
            total=records_to_process,
            report_progress=self._report_progress
        )

        with progress_indicator:
            while records_to_process > 0:
                PyErr_CheckSignals()
                ret = reader.nextRecord()
                if ret == RecordResult.RecordResultAllocationRecord:
                    aggregator.addAllocation(reader.getLatestAllocation())
                    records_to_process -= 1
                    progress_indicator.update(1)
                elif ret == RecordResult.RecordResultMemoryRecord:
                    pass
                else:
                    assert ret != RecordResult.RecordResultMemorySnapshot
                    assert ret != RecordResult.RecordResultAggregatedAllocationRecord
                    break

        for elem in Py_ListFromSnapshotAllocationRecords(
            aggregator.getSnapshotAllocations(merge_threads)
        ):
            alloc = AllocationRecord(elem)
            (<AllocationRecord> alloc)._reader = reader_sp
            yield alloc

        reader.close()

    def get_high_watermark_allocation_records(self, merge_threads=True):
        self._ensure_not_closed()
        if self._header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
            yield from self._reaggregate_allocations(
                self._header["stats"]["n_allocations"],
                merge_threads,
                leaks=False,
            )
            return

        # If allocation 0 caused the peak, we need to process 1 record, etc
        cdef size_t max_records = self._high_watermark.index + 1
        yield from self._aggregate_allocations(max_records, merge_threads)

    def get_leaked_allocation_records(self, merge_threads=True):
        self._ensure_not_closed()
        if self._header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
            yield from self._reaggregate_allocations(
                self._header["stats"]["n_allocations"],
                merge_threads,
                leaks=True,
            )
            return

        cdef size_t max_records = self._header["stats"]["n_allocations"]
        yield from self._aggregate_allocations(max_records, merge_threads)

    def get_temporary_allocation_records(self, merge_threads=True, threshold=1):
        self._ensure_not_closed()
        if self._header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
            raise NotImplementedError(
                "Can't find temporary allocations using a pre-aggregated capture file."
            )

        cdef size_t max_records = self._header["stats"]["n_allocations"]
        yield from self._aggregate_allocations(
            max_records,
            merge_threads,
            temporary_buffer_size=threshold + 1,
        )

    def get_temporal_allocation_records(self, merge_threads=True):
        self._ensure_not_closed()
        if self._header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
            raise NotImplementedError(
                "Can't get allocation history using a pre-aggregated capture file."
            )

        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path))
        )
        cdef RecordReader* reader = reader_sp.get()

        cdef size_t records_to_process = self._header["stats"]["n_allocations"]
        cdef ProgressIndicator progress_indicator = ProgressIndicator(
            "Processing allocation records",
            total=records_to_process,
            report_progress=self._report_progress
        )

        cdef AllocationLifetimeAggregator aggregator
        cdef _Allocation allocation
        cdef size_t memory_records_seen = 0

        with progress_indicator:
            while records_to_process > 0:
                PyErr_CheckSignals()
                ret = reader.nextRecord()
                if ret == RecordResult.RecordResultAllocationRecord:
                    allocation = reader.getLatestAllocation()
                    if merge_threads:
                        allocation.tid = NO_THREAD_INFO
                    aggregator.addAllocation(allocation)
                    records_to_process -= 1
                    progress_indicator.update(1)
                elif ret == RecordResult.RecordResultMemoryRecord:
                    memory_records_seen += 1
                    if self._memory_snapshot_stride and (memory_records_seen - 1) % self._memory_snapshot_stride != 0:
                        continue
                    aggregator.captureSnapshot()
                else:
                    assert ret != RecordResult.RecordResultMemorySnapshot
                    assert ret != RecordResult.RecordResultAggregatedAllocationRecord
                    break

        cdef TemporalAllocationGenerator gen = TemporalAllocationGenerator()
        gen.setup(move(aggregator.generateIndex()), reader_sp)
        yield from gen

    def get_temporal_high_water_mark_allocation_records(self, merge_threads=True):
        self._ensure_not_closed()
        if self._header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
            raise NotImplementedError(
                "Can't get allocation history using a pre-aggregated capture file."
            )

        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path))
        )
        cdef RecordReader* reader = reader_sp.get()

        cdef size_t records_to_process = self._header["stats"]["n_allocations"]
        cdef ProgressIndicator progress_indicator = ProgressIndicator(
            "Processing allocation records",
            total=records_to_process,
            report_progress=self._report_progress
        )

        cdef HighWaterMarkAggregator aggregator
        cdef _Allocation allocation
        cdef size_t memory_records_seen = 0

        with progress_indicator:
            while records_to_process > 0:
                PyErr_CheckSignals()
                ret = reader.nextRecord()
                if ret == RecordResult.RecordResultAllocationRecord:
                    allocation = reader.getLatestAllocation()
                    if merge_threads:
                        allocation.tid = 0
                    aggregator.addAllocation(allocation)
                    records_to_process -= 1
                    progress_indicator.update(1)
                elif ret == RecordResult.RecordResultMemoryRecord:
                    memory_records_seen += 1
                    if self._memory_snapshot_stride and (memory_records_seen - 1) % self._memory_snapshot_stride != 0:
                        continue
                    aggregator.captureSnapshot()
                else:
                    assert ret != RecordResult.RecordResultMemorySnapshot
                    assert ret != RecordResult.RecordResultAggregatedAllocationRecord
                    break

        cdef TemporalAllocationGenerator gen = TemporalAllocationGenerator()
        gen.setup(move(aggregator.generateIndex()), reader_sp)
        hwm_by_snapshot = aggregator.highWaterMarkBytesBySnapshot()
        return gen, hwm_by_snapshot

    def get_allocation_records(self):
        self._ensure_not_closed()
        if self._header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
            raise NotImplementedError(
                "Can't get all allocations from a pre-aggregated capture file."
            )

        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path))
        )
        cdef RecordReader* reader = reader_sp.get()

        while True:
            PyErr_CheckSignals()
            ret = reader.nextRecord()
            if ret == RecordResult.RecordResultAllocationRecord:
                alloc = AllocationRecord(reader.getLatestAllocation().toPythonObject())
                (<AllocationRecord> alloc)._reader = reader_sp
                yield alloc
            elif ret == RecordResult.RecordResultMemoryRecord:
                pass
            elif ret == RecordResult.RecordResultMemorySnapshot:
                pass
            else:
                break

        reader.close()

    def get_memory_snapshots(self):
        for record in self._memory_snapshots:
            yield MemorySnapshot(record.ms_since_epoch, record.rss, record.heap)

    def get_object_lifetime_events(self, included_objects=None):
        """Return all object lifetime events recorded in the file.

        This method yields TrackedObjectRecord instances representing Python objects
        that were tracked during the recording session.

        Args:
            included_objects (list, optional): If provided, only yield
                lifetime events for objects in the list.

        Returns:
            An iterator of TrackedObjectRecord instances.
        """
        self._ensure_not_closed()

        object_ids = set()
        if included_objects is not None:
            object_ids = {id(obj) for obj in included_objects}

        cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(self._path)),
            True,  # track_stacks
            True   # track_object_lifetimes
        )
        cdef RecordReader* reader = reader_sp.get()

        # Check if the file was generated with object lifetime tracking enabled
        cdef object header = reader.getHeader()
        if not header["track_object_lifetimes"]:
            raise RuntimeError(
                "Object lifetime events are not available in this capture file. "
                "The file was not generated with track_object_lifetimes=True."
            )

        cdef TrackedObject tracked_object
        while True:
            PyErr_CheckSignals()
            ret = reader.nextRecord()
            if ret == RecordResult.RecordResultObjectRecord:
                tracked_object = reader.getLatestObject()
                if included_objects is not None and tracked_object.address not in object_ids:
                    continue
                object_record = TrackedObjectRecord(tracked_object.toPythonObject())
                object_record._reader = reader_sp
                yield object_record
            elif ret == RecordResult.RecordResultAllocationRecord:
                pass
            elif ret == RecordResult.RecordResultMemoryRecord:
                pass
            elif ret == RecordResult.RecordResultMemorySnapshot:
                pass
            else:
                break

        reader.close()

    @property
    def metadata(self):
        return _create_metadata(self._header, self._high_watermark.peak_memory)


def compute_statistics(
    file_name,
    *,
    report_progress=False,
    num_largest=5,
):
    cdef shared_ptr[RecordReader] reader_sp = make_shared[RecordReader](
        unique_ptr[FileSource](new FileSource(file_name))
    )
    cdef RecordReader* reader = reader_sp.get()

    cdef header = reader.getHeader()
    total = header["stats"]["n_allocations"] or None

    if header["file_format"] == FileFormat.AGGREGATED_ALLOCATIONS:
        raise NotImplementedError(
            "Can't compute statistics using a pre-aggregated capture file."
        )

    cdef AllocationStatsAggregator aggregator
    cdef ProgressIndicator progress_indicator = ProgressIndicator(
        "Computing statistics",
        total=total,
        report_progress=report_progress,
    )
    with progress_indicator:
        while True:
            PyErr_CheckSignals()
            ret = reader.nextRecord()
            if ret == RecordResult.RecordResultAllocationRecord:
                aggregator.addAllocation(
                    reader.getLatestAllocation(),
                    reader.getLatestPythonLocationId(reader.getLatestAllocation()),
                )
                progress_indicator.update(1)
            elif ret == RecordResult.RecordResultMemoryRecord:
                pass
            elif ret == RecordResult.RecordResultMemorySnapshot:
                pass
            elif ret == RecordResult.RecordResultObjectRecord:
                pass
            else:
                assert ret != RecordResult.RecordResultMemorySnapshot
                assert ret != RecordResult.RecordResultAggregatedAllocationRecord
                break

    # Ignore the n_allocations in the header, use our observed value.
    header["stats"]["n_allocations"] = progress_indicator.num_processed

    # Convert allocation counts by allocator/by size to Python dicts.
    cdef dict tmp = aggregator.allocationCountByAllocator()
    allocation_count_by_allocator = {AllocatorType(k).name: v for k, v in tmp.items()}
    cdef dict allocation_count_by_size = aggregator.allocationCountBySize();

    # Convert top locations by bytes allocated/by allocation count to dicts
    unknown = ("<unknown>", "<unknown>", 0)

    top_locations_by_size = [
        ((reader.Py_GetLocation(size_and_loc.second) or unknown), size_and_loc.first)
        for size_and_loc in aggregator.topLocationsBySize(num_largest)
    ]

    top_locations_by_count = [
        ((reader.Py_GetLocation(count_and_loc.second) or unknown), count_and_loc.first)
        for count_and_loc in aggregator.topLocationsByCount(num_largest)
    ]

    # And we're done!
    cdef uint64_t peak_memory = aggregator.peakBytesAllocated()
    return Stats(
        metadata=_create_metadata(header, peak_memory),
        total_num_allocations=aggregator.totalAllocations(),
        total_memory_allocated=aggregator.totalBytesAllocated(),
        peak_memory_allocated=peak_memory,
        allocation_count_by_size=allocation_count_by_size,
        allocation_count_by_allocator=allocation_count_by_allocator,
        top_locations_by_size=top_locations_by_size,
        top_locations_by_count=top_locations_by_count,
    )


def dump_all_records(object file_name):
    cdef str path = str(file_name)
    if not pathlib.Path(path).exists():
        raise IOError(f"No such file: {path}")

    cdef shared_ptr[RecordReader] _reader = make_shared[RecordReader](
            unique_ptr[FileSource](new FileSource(path)))
    _reader.get().dumpAllRecords()


cdef class SocketReader:
    cdef BackgroundSocketReader* _impl
    cdef shared_ptr[RecordReader] _reader
    cdef object _header
    cdef object _port

    def __cinit__(self, int port):
        self._impl = NULL

    def __init__(self, port: int):
        self._header = {}
        self._port = port

    cdef _teardown(self):
        with nogil:
            del self._impl
        self._impl = NULL

    cdef unique_ptr[SocketSource] _make_source(self) except*:
        # Creating a SocketSource can raise Python exceptions (if is interrupted by signal
        # handlers). If this happens, this method will propagate the appropriate exception.
        # We cannot use make_unique or C++ exceptions from SocketSource() won't be caught.
        cdef SocketSource* source = new SocketSource(self._port)
        return unique_ptr[SocketSource](source)

    def __enter__(self):
        if self._impl is not NULL:
            raise ValueError(
                "Can not enter the context of a SocketReader object more than "
                "once, at the same time."
            )

        self._reader = make_shared[RecordReader](move(self._make_source()))
        self._header = self._reader.get().getHeader()

        self._impl = new BackgroundSocketReader(self._reader)
        self._impl.start()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        assert self._impl is not NULL

        self._teardown()
        self._reader.get().close()

    def __dealloc__(self):
        if self._impl is not NULL:
            self._teardown()

    @property
    def command_line(self):
        if not self._header:
            return None
        return self._header["command_line"]

    @property
    def is_active(self):
        if self._impl == NULL:
            return False
        return self._impl.is_active()

    @property
    def pid(self):
        if not self._header:
            return None
        return self._header["pid"]

    @property
    def has_native_traces(self):
        if not self._header:
            return False
        return self._header["native_traces"]

    def get_current_snapshot(self, *, bool merge_threads):
        if self._impl is NULL:
            return

        snapshot_allocations = self._impl.Py_GetSnapshotAllocationRecords(merge_threads=merge_threads)
        for elem in snapshot_allocations:
            alloc = AllocationRecord(elem)
            (<AllocationRecord> alloc)._reader = self._reader
            yield alloc

cpdef enum SymbolicSupport:
    NONE = 1
    FUNCTION_NAME_ONLY = 2
    TOTAL = 3

def get_symbolic_support():
    locations = unwindHere()
    for location in locations:
        function, file, line = location.split(":")
        if function != "_PyEval_EvalFrameDefault":
            continue
        if not file:
            return SymbolicSupport.FUNCTION_NAME_ONLY
        return SymbolicSupport.TOTAL
    return SymbolicSupport.NONE


cdef extern from "<dlfcn.h>":
    int _RTLD_NOW "RTLD_NOW"
    void* _RTLD_DEFAULT "RTLD_DEFAULT"


RTLD_NOW = _RTLD_NOW
RTLD_DEFAULT = <long long>_RTLD_DEFAULT


cdef extern from "snapshot.h":
    """
    std::vector<memray::tracking_api::AggregatedAllocation>
    collectAllocations(const memray::api::HighWaterMarkAggregator& aggregator)
    {
        std::vector<memray::tracking_api::AggregatedAllocation> ret;
        aggregator.visitAllocations([&](const memray::tracking_api::AggregatedAllocation& agg) {
            if (agg.n_allocations_in_high_water_mark || agg.n_allocations_leaked) {
                ret.push_back(agg);
            }
            return true;
        });
        return ret;
    }
    """
    vector[AggregatedAllocation] collectAllocations(HighWaterMarkAggregator) except+


cdef class HighWaterMarkAggregatorTestHarness:
    cdef HighWaterMarkAggregator aggregator

    def add_allocation(
        self,
        tid,
        address,
        size,
        allocator,
        native_frame_id,
        frame_index,
        native_segment_generation,
    ):
        cdef _Allocation allocation
        allocation.tid = tid
        allocation.address = address
        allocation.size = size
        allocation.allocator = <Allocator><int>allocator
        allocation.native_frame_id = native_frame_id
        allocation.frame_index = frame_index
        allocation.native_segment_generation = native_segment_generation
        allocation.n_allocations = 1
        self.aggregator.addAllocation(allocation)

    def capture_snapshot(self):
        return self.aggregator.captureSnapshot()

    def high_water_mark_bytes_by_snapshot(self):
        return self.aggregator.highWaterMarkBytesBySnapshot()

    def get_current_heap_size(self):
        return self.aggregator.getCurrentHeapSize()

    def get_temporal_allocations(self):
        cdef shared_ptr[RecordReader] reader
        cdef TemporalAllocationGenerator gen = TemporalAllocationGenerator()
        gen.setup(move(self.aggregator.generateIndex()), reader)
        yield from gen

    def get_allocations(self):
        ret = []
        for allocation in collectAllocations(self.aggregator):
            ret.append(
                dict(
                    tid=allocation.tid,
                    allocator=AllocatorType(<int>allocation.allocator),
                    native_frame_id=allocation.native_frame_id,
                    frame_index=allocation.frame_index,
                    native_segment_generation=allocation.native_segment_generation,
                    n_allocations_in_high_water_mark=allocation.n_allocations_in_high_water_mark,
                    n_allocations_leaked=allocation.n_allocations_leaked,
                    bytes_in_high_water_mark=allocation.bytes_in_high_water_mark,
                    bytes_leaked=allocation.bytes_leaked,
                )
            )
        return ret


cdef class AllocationLifetimeAggregatorTestHarness:
    cdef AllocationLifetimeAggregator aggregator

    def add_allocation(
        self,
        tid,
        address,
        size,
        allocator,
        native_frame_id,
        frame_index,
        native_segment_generation,
    ):
        cdef _Allocation allocation
        allocation.tid = tid
        allocation.address = address
        allocation.size = size
        allocation.allocator = <Allocator><int>allocator
        allocation.native_frame_id = native_frame_id
        allocation.frame_index = frame_index
        allocation.native_segment_generation = native_segment_generation
        allocation.n_allocations = 1
        self.aggregator.addAllocation(allocation)

    def capture_snapshot(self):
        return self.aggregator.captureSnapshot()

    def get_allocations(self):
        cdef shared_ptr[RecordReader] reader
        cdef TemporalAllocationGenerator gen = TemporalAllocationGenerator()
        gen.setup(move(self.aggregator.generateIndex()), reader)
        yield from gen


cdef class RecordWriterTestHarness:
    """A Python wrapper around the C++ RecordWriter class for testing purposes."""

    cdef unique_ptr[RecordWriter] _writer

    def __cinit__(
        self,
        str file_path,
        bool native_traces=False,
        bool trace_python_allocators=False,
        track_object_lifetimes=False,
        records.FileFormat file_format=records.FileFormat.ALL_ALLOCATIONS,
        records.thread_id_t main_tid=1,
        size_t skipped_frames=0,
        str command_line="memray test harness",
    ):
        """Initialize a new RecordWriterTestHarness.

        Args:
            file_path: Path to the output file
            native_traces: Whether to include native traces
            trace_python_allocators: Whether to trace Python allocators
            track_object_lifetimes: Whether to track Python object lifetimes
            file_format: The format of the output file
        """
        self._writer = createRecordWriter(
            unique_ptr[Sink](new FileSink(file_path.encode('utf-8'), True, False)),
            command_line.encode(),
            native_traces,
            file_format,
            trace_python_allocators,
            track_object_lifetimes,
        )
        self._writer.get().setMainTidAndSkippedFrames(main_tid, skipped_frames)
        self.write_header(False)

    def write_header(self, bool seek_to_start) -> None:
        if not self._writer.get().writeHeader(seek_to_start):
            raise RuntimeError("Failed to write header")

    def write_memory_record(self, uint64_t ms_since_epoch, size_t rss) -> bool:
        """Write a memory record to the file."""
        cdef records.MemoryRecord record
        record.ms_since_epoch = ms_since_epoch
        record.rss = rss
        return self._writer.get().writeRecord(record)

    def write_code_object(
        self,
        records.code_object_id_t id,
        str function_name,
        str filename,
        bytes linetable,
        int firstlineno,
    ) -> bool:
        """Write a code object record to the file."""
        return self._writer.get().writeRecord(
            pair[records.code_object_id_t, records.CodeObjectInfo](
                id,
                records.CodeObjectInfo(
                    function_name.encode(),
                    filename.encode(),
                    linetable,
                    firstlineno,
                )
            )
        )

    @staticmethod
    def get_linetable(code_object):
        """Get the linetable from a code object."""
        cdef const char* addr
        cdef size_t size
        import types
        if not isinstance(code_object, types.CodeType):
            raise TypeError("Expected a code object")
        addr = codeGetLinetable(<PyCodeObject*>code_object, &size)
        return <bytes>addr[:size]

    @staticmethod
    def get_lasti(frame_object):
        """Get the linetable from a code object."""
        import types
        if not isinstance(frame_object, types.FrameType):
            raise TypeError("Expected a frame object")
        return frameGetLasti(<PyFrameObject*>frame_object)

    def write_unresolved_native_frame(self, uintptr_t ip, size_t index) -> bool:
        """Write an unresolved native frame record to the file."""
        cdef records.UnresolvedNativeFrame record
        record.ip = ip
        record.index = index
        return self._writer.get().writeRecord(record)

    def write_allocation_record(self, records.thread_id_t tid, uintptr_t address,
                                     size_t size, unsigned char allocator,
                                     size_t native_frame_id=0) -> bool:
        """Write a native allocation record to the file."""
        cdef records.AllocationRecord record
        record.address = address
        record.size = size
        record.allocator = <records.Allocator>allocator
        record.native_frame_id = native_frame_id
        return self._writer.get().writeThreadSpecificRecord(tid, record)

    def write_frame_push(
        self,
        records.thread_id_t tid,
        records.code_object_id_t code_object_id,
        int instruction_offset,
        bool is_entry_frame,
    ) -> bool:
        """Write a frame push record to the file."""
        cdef records.FramePush record
        record.frame.code_object_id = code_object_id
        record.frame.instruction_offset = instruction_offset
        record.frame.is_entry_frame = is_entry_frame
        return self._writer.get().writeThreadSpecificRecord(tid, record)

    def write_frame_pop(self, records.thread_id_t tid, size_t count) -> bool:
        """Write a frame pop record to the file."""
        cdef records.FramePop record
        record.count = count
        return self._writer.get().writeThreadSpecificRecord(tid, record)

    def write_thread_record(self, records.thread_id_t tid, str name) -> bool:
        """Write a thread record to the file."""
        cdef records.ThreadRecord record
        cdef bytes name_bytes = name.encode('utf-8')
        record.name = name_bytes
        return self._writer.get().writeThreadSpecificRecord(tid, record)

    def write_mappings(self, list mappings) -> bool:
        """Write memory mappings to the file."""
        return self._writer.get().writeMappings(
            [
                records.ImageSegments(
                    mapping['filename'].encode('utf-8'),
                    mapping['addr'],
                    [
                        records.Segment(seg["vaddr"], seg["memsz"])
                        for seg in mapping["segments"]
                    ],
                )
                for mapping in mappings
            ]
        )

    def write_object_record(
        self,
        records.thread_id_t tid,
        uintptr_t address,
        bool is_created,
        records.frame_id_t native_frame_id = 0,
    ) -> bool:
        """Write an object tracking record to the file."""
        cdef records.ObjectRecord record
        record.address = address
        record.is_created = is_created
        record.native_frame_id = native_frame_id
        return self._writer.get().writeThreadSpecificRecord(tid, record)

    def write_trailer(self) -> bool:
        """Write the trailer to the file."""
        return self._writer.get().writeTrailer()



import html
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import TextIO

from memray import AllocationRecord
from memray import AllocatorType
from memray import MemorySnapshot
from memray import Metadata
from memray.reporters.common import format_thread_name
from memray.reporters.templates import render_report


class TableReporter:
    def __init__(
        self,
        data: List[Dict[str, Any]],
        *,
        memory_records: Iterable[MemorySnapshot],
    ):
        super().__init__()
        self.data = data
        self.memory_records = memory_records

    @classmethod
    def from_snapshot(
        cls,
        allocations: Iterable[AllocationRecord],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        **kwargs: Any,
    ) -> "TableReporter":
        result = []
        for record in allocations:
            stack_trace = (
                list(record.hybrid_stack_trace(max_stacks=1))
                if native_traces
                else record.stack_trace(max_stacks=1)
            )
            stack = "???"
            if stack_trace:
                function, file, line = stack_trace[0]
                stack = f"{function} at {file}:{line}"

            allocator = AllocatorType(record.allocator)
            result.append(
                {
                    "tid": format_thread_name(record),
                    "size": record.size,
                    "allocator": allocator.name.lower(),
                    "n_allocations": record.n_allocations,
                    "stack_trace": html.escape(stack),
                }
            )

        return cls(result, memory_records=memory_records)

    def render(
        self,
        outfile: TextIO,
        metadata: Metadata,
        show_memory_leaks: bool,
        merge_threads: bool,
        inverted: bool,
        no_web: bool = False,
    ) -> None:
        if not merge_threads:
            raise NotImplementedError("TableReporter only supports merged threads.")
        if inverted:
            raise NotImplementedError(
                "TableReporter does not support inverted argument"
            )
        html_code = render_report(
            kind="table",
            data=self.data,
            metadata=metadata,
            memory_records=self.memory_records,
            show_memory_leaks=show_memory_leaks,
            merge_threads=merge_threads,
            inverted=inverted,
            no_web=no_web,
        )
        print(html_code, file=outfile)



import pathlib
import typing
from dataclasses import dataclass


@dataclass(frozen=True)
class Destination:
    pass


@dataclass(frozen=True)
class FileDestination(Destination):
    """Specify an output file to write captured allocations into.

    Args:
        path: The path to the output file.
        overwrite: By default, if a file already exists at that path an
            exception will be raised. If you provide ``overwrite=True``, then
            the existing file will be overwritten instead.
    """

    path: typing.Union[pathlib.Path, str]
    overwrite: bool = False
    compress_on_exit: bool = True


@dataclass(frozen=True)
class SocketDestination(Destination):
    """Specify a port to serve captured allocations on.

    When a ``SocketDestination`` is passed to the `Tracker` constructor, the
    process will immediately create a server socket on the given port and wait
    for a reader to connect (see :ref:`Live Tracking`). The `Tracker`
    constructor will not return until a client has connected. Any records the
    tracker goes on to capture will be written over the socket to the attached
    client.

    Args:
        server_port: The port to accept a client connection on.
        address: The address to bind the server socket to. This should
            generally be left alone, but you might want to use ``"0.0.0.0"`` to
            accept connections from clients on other machines. Note that
            sending records to clients on other machines is generally a bad
            idea, though. In particular, this won't play nicely with
            :ref:`Native Tracking`, because the client on the remote machine
            won't have access to the shared libraries used by the tracked
            process.
    """

    server_port: int
    address: str = "127.0.0.1"



from typing import Any


class MemrayError(Exception):
    """Exceptions raised in this package."""


class MemrayCommandError(MemrayError):
    """Exceptions raised from this package's CLI commands."""

    def __init__(self, *args: Any, exit_code: int) -> None:
        super().__init__(*args)
        self.exit_code = exit_code



import collections
import html
import itertools
import linecache
import sys
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import TextIO
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from memray import AllocationRecord
from memray import MemorySnapshot
from memray import Metadata
from memray._memray import Interval
from memray._memray import TemporalAllocationRecord
from memray.reporters.common import format_thread_name
from memray.reporters.frame_tools import StackFrame
from memray.reporters.frame_tools import is_cpython_internal
from memray.reporters.frame_tools import is_frame_from_import_system
from memray.reporters.frame_tools import is_frame_interesting
from memray.reporters.templates import render_report

PythonStackElement = Tuple[str, str, int]
MAX_STACKS = int(sys.getrecursionlimit() // 2.5)

T = TypeVar("T")


class RecordData(TypedDict):
    thread_name: str
    size: Optional[int]
    n_allocations: Optional[int]
    intervals: Optional[List[Interval]]


class FrameNodeDict(TypedDict):
    name: str
    location: Tuple[str, str, int]
    value: int
    children: List[int]
    n_allocations: int
    interesting: bool
    thread_id: str
    import_system: bool


class FlameGraphNodeDict(TypedDict):
    name: List[int]
    function: List[int]
    filename: List[int]
    lineno: List[int]
    children: List[List[int]]
    value: List[int]
    n_allocations: List[int]
    thread_id: List[int]
    interesting: List[int]
    import_system: List[int]


def create_framegraph_node_from_stack_frame(
    stack_frame: StackFrame,
    thread_id: str,
    import_system: bool = False,
) -> FrameNodeDict:
    function, filename, lineno = stack_frame

    name = (
        # Use the source file line.
        linecache.getline(filename, lineno)
        # Or just describe where it is from
        or f"{function} at {filename}:{lineno}"
    )
    return {
        "name": name,
        "location": (html.escape(function), html.escape(filename), lineno),
        "value": 0,
        "children": [],
        "n_allocations": 0,
        "interesting": (
            is_frame_interesting(stack_frame)
            and not is_frame_from_import_system(stack_frame)
        ),
        "thread_id": thread_id,
        "import_system": import_system,
    }


class StringRegistry:
    def __init__(self) -> None:
        self.strings: List[str] = []
        self._index_by_string: Dict[str, int] = {}

    def register(self, string: str) -> int:
        idx = self._index_by_string.setdefault(string, len(self.strings))
        if idx == len(self.strings):
            self.strings.append(string)
        return idx


class FlameGraphReporter:
    def __init__(
        self,
        data: Dict[str, Any],
        *,
        memory_records: Iterable[MemorySnapshot],
    ) -> None:
        super().__init__()
        self.data = data
        self.memory_records = memory_records

    @classmethod
    def generate_nodes(
        cls,
        frames: List[FrameNodeDict],
        all_strings: StringRegistry,
        temporal: bool,
    ) -> FlameGraphNodeDict:
        nodes = cast(FlameGraphNodeDict, collections.defaultdict(list))
        for frame in frames:
            nodes["name"].append(all_strings.register(frame["name"]))
            nodes["function"].append(all_strings.register(frame["location"][0]))
            nodes["filename"].append(all_strings.register(frame["location"][1]))
            nodes["lineno"].append(frame["location"][2])
            nodes["children"].append(frame["children"])
            if not temporal:
                nodes["value"].append(frame["value"])
                nodes["n_allocations"].append(frame["n_allocations"])
            nodes["thread_id"].append(all_strings.register(frame["thread_id"]))
            nodes["interesting"].append(int(frame["interesting"]))
            nodes["import_system"].append(int(frame["import_system"]))
        return nodes

    @classmethod
    def generate_frames(
        cls,
        stack_it: Iterable[Tuple[int, PythonStackElement]],
        frames: List[FrameNodeDict],
        node_index_by_key: Dict[Tuple[int, StackFrame, str], int],
        record: RecordData,
        inverted: bool,
        interval_list: List[Tuple[int, Optional[int], int, int, int]],
    ) -> None:
        current_frame_id = 0
        current_frame = frames[0]

        if record["size"] is not None:
            current_frame["value"] += record["size"]
        if record["n_allocations"] is not None:
            current_frame["n_allocations"] += record["n_allocations"]

        num_skipped_frames = 0
        is_import_system = False

        for index, stack_frame in stack_it:
            node_key = (current_frame_id, stack_frame, record["thread_name"])
            if node_key not in node_index_by_key:
                if is_cpython_internal(stack_frame):
                    num_skipped_frames += 1
                    continue
                # update import_system only if we're generting normal flamegraph
                if not inverted and not is_import_system:
                    is_import_system = is_frame_from_import_system(stack_frame)

                new_node_id = len(frames)
                node_index_by_key[node_key] = new_node_id
                current_frame["children"].append(new_node_id)
                frames.append(
                    create_framegraph_node_from_stack_frame(
                        stack_frame,
                        import_system=is_import_system,
                        thread_id=record["thread_name"],
                    )
                )
            current_frame_id = node_index_by_key[node_key]
            current_frame = frames[current_frame_id]
            is_import_system = current_frame["import_system"]

            if record["size"] is not None:
                current_frame["value"] += record["size"]
            if record["n_allocations"] is not None:
                current_frame["n_allocations"] += record["n_allocations"]

            if index - num_skipped_frames > MAX_STACKS:
                current_frame["name"] = "<STACK TOO DEEP>"
                current_frame["location"] = ("...", "...", 0)
                break

        if record["intervals"] is not None:
            interval_list.extend(
                (
                    interval.allocated_before_snapshot,
                    interval.deallocated_before_snapshot,
                    current_frame_id,
                    interval.n_allocations,
                    interval.n_bytes,
                )
                for interval in record["intervals"]
            )

    @classmethod
    def _create_root_node(cls) -> FrameNodeDict:
        return {
            "name": "<root>",
            "location": (html.escape("<tracker>"), "<b>memray</b>", 0),
            "value": 0,
            "children": [],
            "n_allocations": 0,
            "thread_id": "0x0",
            "interesting": True,
            "import_system": False,
        }

    @classmethod
    def _drop_import_system_frames(
        cls,
        stack: Sequence[PythonStackElement],
    ) -> Iterable[PythonStackElement]:
        return reversed(
            list(
                itertools.takewhile(
                    lambda e: not is_frame_from_import_system(e),
                    reversed(stack),
                )
            )
        )

    @classmethod
    def _from_any_snapshot(
        cls,
        allocations: Iterable[Union[AllocationRecord, TemporalAllocationRecord]],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        temporal: bool,
        inverted: Optional[bool] = None,
    ) -> "FlameGraphReporter":
        inverted = False if inverted is None else inverted

        frames: List[FrameNodeDict] = [cls._create_root_node()]
        inverted_no_imports_frames: List[FrameNodeDict] = []

        if inverted:
            inverted_no_imports_frames = [cls._create_root_node()]

        interval_list: List[Tuple[int, Optional[int], int, int, int]] = []
        no_imports_interval_list: List[Tuple[int, Optional[int], int, int, int]] = []

        NodeKey = Tuple[int, StackFrame, str]
        node_index_by_key: Dict[NodeKey, int] = {}
        inverted_no_imports_node_index_by_key: Dict[NodeKey, int] = {}

        unique_threads: Set[str] = set()
        for record in allocations:
            unique_threads.add(format_thread_name(record))

            record_data: RecordData
            if temporal:
                assert isinstance(record, TemporalAllocationRecord)
                record_data = {
                    "thread_name": format_thread_name(record),
                    "intervals": record.intervals,
                    "size": None,
                    "n_allocations": None,
                }
            else:
                assert not isinstance(record, TemporalAllocationRecord)
                record_data = {
                    "thread_name": format_thread_name(record),
                    "intervals": None,
                    "size": record.size,
                    "n_allocations": record.n_allocations,
                }

            stack = (
                tuple(record.hybrid_stack_trace())
                if native_traces
                else record.stack_trace()
            )

            if not inverted:
                # normal flamegraph
                cls.generate_frames(
                    stack_it=enumerate(reversed(stack)),
                    frames=frames,
                    node_index_by_key=node_index_by_key,
                    record=record_data,
                    inverted=inverted,
                    interval_list=interval_list,
                )
            else:
                # inverted flamegraph tree with all nodes
                cls.generate_frames(
                    stack_it=enumerate(stack),
                    frames=frames,
                    node_index_by_key=node_index_by_key,
                    record=record_data,
                    inverted=inverted,
                    interval_list=interval_list,
                )

                # inverted flamegraph tree without import system nodes
                cls.generate_frames(
                    stack_it=enumerate(cls._drop_import_system_frames(stack)),
                    frames=inverted_no_imports_frames,
                    node_index_by_key=inverted_no_imports_node_index_by_key,
                    record=record_data,
                    inverted=inverted,
                    interval_list=no_imports_interval_list,
                )

        all_strings = StringRegistry()
        nodes = cls.generate_nodes(
            frames=frames, all_strings=all_strings, temporal=temporal
        )
        inverted_no_imports_nodes = cls.generate_nodes(
            frames=inverted_no_imports_frames,
            all_strings=all_strings,
            temporal=temporal,
        )

        data = {
            "unique_threads": tuple(
                all_strings.register(t) for t in sorted(unique_threads)
            ),
            "nodes": nodes,
            "inverted_no_imports_nodes": inverted_no_imports_nodes,
            "strings": all_strings.strings,
        }

        if temporal:
            data["intervals"] = interval_list
            data["no_imports_interval_list"] = no_imports_interval_list

        return cls(data, memory_records=memory_records)

    @classmethod
    def from_snapshot(
        cls,
        allocations: Iterable[AllocationRecord],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        inverted: Optional[bool] = None,
    ) -> "FlameGraphReporter":
        return cls._from_any_snapshot(
            allocations,
            memory_records=memory_records,
            native_traces=native_traces,
            temporal=False,
            inverted=inverted,
        )

    @classmethod
    def from_temporal_snapshot(
        cls,
        allocations: Iterable[TemporalAllocationRecord],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        high_water_mark_by_snapshot: Optional[List[int]],
        inverted: Optional[bool] = None,
    ) -> "FlameGraphReporter":
        ret = cls._from_any_snapshot(
            allocations,
            memory_records=memory_records,
            native_traces=native_traces,
            temporal=True,
            inverted=inverted,
        )
        ret.data["high_water_mark_by_snapshot"] = high_water_mark_by_snapshot
        return ret

    def render(
        self,
        outfile: TextIO,
        metadata: Metadata,
        show_memory_leaks: bool,
        merge_threads: bool,
        inverted: bool,
        no_web: bool = False,
    ) -> None:
        kind = "temporal_flamegraph" if "intervals" in self.data else "flamegraph"
        html_code = render_report(
            kind=kind,
            data=self.data,
            metadata=metadata,
            memory_records=self.memory_records,
            show_memory_leaks=show_memory_leaks,
            merge_threads=merge_threads,
            inverted=inverted,
            no_web=no_web,
        )
        print(html_code, file=outfile)



import argparse
import os
from pathlib import Path
from textwrap import dedent

from memray import FileReader
from memray._errors import MemrayCommandError
from memray.commands.common import warn_if_file_is_not_aggregated_and_is_too_big
from memray.commands.common import warn_if_not_enough_symbols
from memray.reporters.summary import SummaryReporter


class SummaryCommand:
    """Generate a terminal-based summary report of the functions that allocate most memory"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("results", help="Results of the tracker run")
        parser.add_argument(
            "-s",
            "--sort-column",
            help="Column number to sort on",
            type=int,
            default=1,
        )
        parser.add_argument(
            "-r",
            "--max-rows",
            help="Maximum number of rows to display",
            type=int,
            default=None,
        )
        alloc_type_group = parser.add_mutually_exclusive_group()
        alloc_type_group.add_argument(
            "--temporary-allocation-threshold",
            metavar="N",
            help=dedent(
                """
                Report temporary allocations, as opposed to leaked allocations
                or high watermark allocations.  An allocation is considered
                temporary if at most N other allocations occur before it is
                deallocated.  With N=0, an allocation is temporary only if it
                is immediately deallocated before any other allocation occurs.
                """
            ),
            action="store",
            dest="temporary_allocation_threshold",
            type=int,
            default=-1,
        )
        alloc_type_group.add_argument(
            "--temporary-allocations",
            help="Equivalent to --temporary-allocation-threshold=1",
            action="store_const",
            dest="temporary_allocation_threshold",
            const=1,
        )

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        max_cols = SummaryReporter.N_COLUMNS
        if args.sort_column < 1 or args.sort_column > max_cols:
            parser.error(f"The --sort-column argument must be between 1 and {max_cols}")

        result_path = Path(args.results)
        if not result_path.exists() or not result_path.is_file():
            raise MemrayCommandError(f"No such file: {args.results}", exit_code=1)

        try:
            reader = FileReader(os.fspath(args.results), report_progress=True)
            if reader.metadata.has_native_traces:
                warn_if_not_enough_symbols()

            if not args.temporary_allocation_threshold >= 0:
                warn_if_file_is_not_aggregated_and_is_too_big(reader, result_path)

            if args.temporary_allocation_threshold >= 0:
                snapshot = iter(
                    reader.get_temporary_allocation_records(
                        threshold=args.temporary_allocation_threshold,
                        merge_threads=False,
                    )
                )
            else:
                snapshot = iter(
                    reader.get_high_watermark_allocation_records(merge_threads=True)
                )
        except OSError as e:
            raise MemrayCommandError(
                f"Failed to parse allocation records in {result_path}\nReason: {e}",
                exit_code=1,
            )
        reporter = SummaryReporter.from_snapshot(
            snapshot,
            native=reader.metadata.has_native_traces,
        )
        reporter.render(sort_column=args.sort_column, max_rows=args.max_rows)



"""Utilities used only by memray's test suite.

If you make changes to this file that move functions around, you will need to
change line numbers in test files as well.
"""
import sys

from posix.mman cimport MAP_ANONYMOUS
from posix.mman cimport MAP_FAILED
from posix.mman cimport MAP_SHARED
from posix.mman cimport PROT_WRITE
from posix.mman cimport mmap
from posix.mman cimport munmap
from posix.unistd cimport read
from posix.unistd cimport write

from _memray.alloc cimport PyMem_Calloc
from _memray.alloc cimport PyMem_Free
from _memray.alloc cimport PyMem_Malloc
from _memray.alloc cimport PyMem_RawCalloc
from _memray.alloc cimport PyMem_RawFree
from _memray.alloc cimport PyMem_RawMalloc
from _memray.alloc cimport PyMem_RawRealloc
from _memray.alloc cimport PyMem_Realloc
from _memray.alloc cimport PyObject_Calloc
from _memray.alloc cimport PyObject_Free
from _memray.alloc cimport PyObject_Malloc
from _memray.alloc cimport PyObject_Realloc
from _memray.alloc cimport aligned_alloc
from _memray.alloc cimport calloc
from _memray.alloc cimport free
from _memray.alloc cimport malloc
from _memray.alloc cimport memalign
from _memray.alloc cimport posix_memalign
from _memray.alloc cimport pvalloc
from _memray.alloc cimport realloc
from _memray.alloc cimport valloc
from _memray.pthread cimport pthread_create
from _memray.pthread cimport pthread_join
from _memray.pthread cimport pthread_t
from cpython.pylifecycle cimport Py_FinalizeEx
from libc.errno cimport errno
from libc.stdint cimport uintptr_t
from libc.stdlib cimport exit as _exit
from libcpp.vector cimport vector

from ._destination import Destination


cdef extern from *:
    """
    #ifdef __linux__
    # include <sys/prctl.h>
    inline int set_thread_name_impl(const char* new_name)
    {
        return prctl(PR_SET_NAME, new_name, NULL, NULL, NULL);
    }
    #else
    # include <errno.h>
    inline int set_thread_name_impl(const char* new_name)
    {
        errno = ENOTSUP;
        return -1;
    }
    #endif
    """
    int set_thread_name_impl(const char* new_name)


def set_thread_name(new_name):
    return set_thread_name_impl(new_name)


cdef class MemoryAllocator:
    cdef void* ptr

    def __cinit__(self):
        self.ptr = NULL

    def free(self):
        if self.ptr == NULL:
            raise RuntimeError("Pointer cannot be NULL")
        free(self.ptr)
        self.ptr = NULL

    def malloc(self, size_t size):
        self.ptr = malloc(size)
        return self.ptr != NULL

    def calloc(self, size_t size):
        self.ptr = calloc(1, size)
        return self.ptr != NULL

    def realloc(self, size_t size):
        self.ptr = malloc(1)
        self.ptr = realloc(self.ptr, size)
        return self.ptr != NULL

    def posix_memalign(self, size_t size):
        rc = posix_memalign(&self.ptr, sizeof(void*), size)
        return rc == 0 and self.ptr != NULL

    def aligned_alloc(self, size_t size):
        self.ptr = aligned_alloc(sizeof(void*), size)
        return self.ptr != NULL

    def memalign(self, size_t size):
        self.ptr = memalign(sizeof(void*), size)
        return self.ptr != NULL

    def valloc(self, size_t size):
        self.ptr = valloc(size)
        return self.ptr != NULL

    def pvalloc(self, size_t size):
        self.ptr = pvalloc(size)
        return self.ptr != NULL

    def run_in_pthread(self, callback):
        cdef pthread_t thread
        cdef int ret = pthread_create(&thread, NULL, &_pthread_worker, <void*>callback)
        if ret != 0:
            raise RuntimeError("Failed to create thread")
        with nogil:
            pthread_join(thread, NULL)


cpdef enum PymallocDomain:
    PYMALLOC_RAW = 1
    PYMALLOC_MEM = 2
    PYMALLOC_OBJECT = 3


cdef class PymallocMemoryAllocator:
    cdef void* ptr
    cdef PymallocDomain domain

    def __cinit__(self, PymallocDomain domain):
        self.ptr = NULL
        self.domain = domain

    def free(self):
        if self.ptr == NULL:
            raise RuntimeError("Pointer cannot be NULL")
        if self.domain == PYMALLOC_RAW:
            PyMem_RawFree(self.ptr)
        elif self.domain == PYMALLOC_MEM:
            PyMem_Free(self.ptr)
        elif self.domain == PYMALLOC_OBJECT:
            PyObject_Free(self.ptr)
        else:
            raise RuntimeError("Invlid pymalloc domain")
        self.ptr = NULL

    def malloc(self, size_t size):
        if self.domain == PYMALLOC_RAW:
            self.ptr = PyMem_RawMalloc(size)
        elif self.domain == PYMALLOC_MEM:
            self.ptr = PyMem_Malloc(size)
        elif self.domain == PYMALLOC_OBJECT:
            self.ptr = PyObject_Malloc(size)
        else:
            raise RuntimeError("Invlid pymalloc domain")

        return self.ptr != NULL

    def calloc(self, size_t size):
        if self.domain == PYMALLOC_RAW:
            self.ptr = PyMem_RawCalloc(1, size)
        elif self.domain == PYMALLOC_MEM:
            self.ptr = PyMem_Calloc(1, size)
        elif self.domain == PYMALLOC_OBJECT:
            self.ptr = PyObject_Calloc(1, size)
        else:
            raise RuntimeError("Invlid pymalloc domain")

        return self.ptr != NULL

    def realloc(self, size_t size):
        if self.domain == PYMALLOC_RAW:
            self.ptr = PyMem_RawRealloc(self.ptr, size)
        elif self.domain == PYMALLOC_MEM:
            self.ptr = PyMem_Realloc(self.ptr, size)
        elif self.domain == PYMALLOC_OBJECT:
            self.ptr = PyObject_Realloc(self.ptr, size)
        else:
            raise RuntimeError("Invlid pymalloc domain")

        return self.ptr != NULL

cdef do_not_optimize_ptr(void* ptr):
    return ptr == <void*>(1)

def _cython_nested_allocation(allocator_fn, size):
    allocator_fn(size)
    cdef void* p = valloc(size);
    do_not_optimize_ptr(p)
    free(p)

cdef class MmapAllocator:
    cdef uintptr_t _address

    def __cinit__(self, size, address=0):
        cdef uintptr_t start_address = address

        self._address = <uintptr_t>mmap(<void *>start_address, size, PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0)
        if <void *>self._address == MAP_FAILED:
            raise MemoryError

    @property
    def address(self):
        return self._address

    def munmap(self, length, offset=0):
        cdef uintptr_t addr = self._address + <uintptr_t> offset
        cdef int ret = munmap(<void *>addr, length)
        if ret != 0:
            raise MemoryError(f"munmap rcode: {ret} errno: {errno}")

cdef void* _pthread_worker(void* arg) noexcept with gil:
    (<object> arg)()

def _cython_allocate_in_two_places(size_t size):
    cdef void* a = allocation_place_a(size)
    do_not_optimize_ptr(a)
    cdef void* b = allocation_place_b(size)
    do_not_optimize_ptr(b)
    free(a)
    free(b)

cdef void* allocation_place_a(size_t size):
    return valloc(size)

cdef void* allocation_place_b(size_t size):
    return valloc(size)

def function_caller(func):
    func()

def allocate_without_gil_held(int wake_up_main_fd, int wake_up_thread_fd):
    cdef char buf = 0
    cdef int write_rc = 0
    cdef int read_rc = 0
    cdef void* p = NULL
    with nogil:
        while write_rc != 1:
            write_rc = write(wake_up_main_fd, &buf, 1)
        while read_rc != 1:
            read_rc = read(wake_up_thread_fd, &buf, 1)
        p = valloc(1234)
    do_not_optimize_ptr(p)
    free(p)
    p = valloc(4321)
    do_not_optimize_ptr(p)
    free(p)

def allocate_cpp_vector(size_t size):
    cdef vector[int] v;
    cdef size_t nelems = <size_t>(size / sizeof(int))
    v.reserve(nelems)
    return v.size()


def fill_cpp_vector(size_t size):
    cdef vector[int] v
    cdef size_t nelems = <size_t>(size / sizeof(int))
    for i in range(nelems):
        v.push_back(i)
    return v.size()

def exit(bint py_finalize=False):
    if py_finalize:
        Py_FinalizeEx()
        _exit(0)
    else:
        with nogil:
            _exit(0)


cdef class PrimeCaches:
    cdef object old_profile
    def __enter__(self):
        self.old_profile = sys.getprofile()
        sys.setprofile(lambda *args: None)
        return self
    def __exit__(self, *args):
        sys.setprofile(self.old_profile)



import argparse
import os
from pathlib import Path
from typing import Optional

from memray._errors import MemrayCommandError
from memray._memray import compute_statistics
from memray.reporters.stats import StatsReporter


class StatsCommand:
    """Generate high level stats of the memory usage in the terminal"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("results", help="Results of the tracker run")

        def valid_positive_int(value: str) -> int:
            try:
                ivalue = int(value)
                if ivalue <= 0:
                    raise ValueError
            except ValueError:
                raise argparse.ArgumentTypeError(
                    f"{value} is an invalid positive int value"
                )

            return ivalue

        parser.add_argument(
            "-n",
            "--num-largest",
            help="Displays the top 'n' largest allocating functions. Default is 5",
            type=valid_positive_int,
            default=5,
        )

        parser.add_argument(
            "--json",
            help="Exports stats to a JSON file",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "-o",
            "--output",
            help="Output file name for JSON output",
            default=None,
        )
        parser.add_argument(
            "-f",
            "--force",
            help="If the JSON output file already exists, overwrite it",
            action="store_true",
            default=False,
        )

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        result_path = Path(args.results)
        if not result_path.exists() or not result_path.is_file():
            raise MemrayCommandError(f"No such file: {args.results}", exit_code=1)
        try:
            stats = compute_statistics(
                os.fspath(args.results),
                report_progress=True,
                num_largest=args.num_largest,
            )
        except OSError as e:
            raise MemrayCommandError(
                f"Failed to compute statistics for {result_path}\nReason: {e}",
                exit_code=1,
            )

        json_output_file: Optional[Path] = None
        if args.json:
            if args.output:
                json_output_file = Path(args.output)
            else:
                filename = str(result_path.name) + ".json"
                if filename.startswith("memray-"):
                    filename = filename[len("memray-") :]
                filename = "memray-stats-" + filename
                json_output_file = result_path.with_name(filename)

            if not args.force and json_output_file.exists():
                raise MemrayCommandError(
                    f"File already exists, will not overwrite: {json_output_file}",
                    exit_code=1,
                )

        reporter = StatsReporter(stats, args.num_largest)
        reporter.render(json_output_file=json_output_file)
        if json_output_file is not None:
            print(f"Wrote {json_output_file}")



import argparse
from contextlib import suppress
from typing import Optional

from memray import SocketReader
from memray._errors import MemrayCommandError
from memray.reporters.tui import TUIApp


class LiveCommand:
    """Remotely monitor allocations in a text-based interface"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "port",
            help="Remote port to connect to",
            default=None,
            type=int,
        )

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        with suppress(KeyboardInterrupt):
            self.start_live_interface(args.port)

    def start_live_interface(
        self, port: int, cmdline_override: Optional[str] = None
    ) -> None:
        if port >= 2**16 or port <= 0:
            raise MemrayCommandError(f"Invalid port: {port}", exit_code=1)
        with SocketReader(port=port) as reader:
            TUIApp(reader, cmdline_override=cmdline_override).run()



"""Tools for processing and filtering stack frames."""
import functools
import re
from typing import Tuple

Symbol = str
File = str
Lineno = int
StackFrame = Tuple[Symbol, File, Lineno]

RE_CPYTHON_PATHS = re.compile(r"(Include|Objects|Modules|Python|cpython).*\.[c|h]$")

SYMBOL_IGNORELIST = {
    "PyObject_Call",
    "call_function",
    "classmethoddescr_call",
    "cmpwrapper_call",
    "do_call_core",
    "fast_function",
    "function_call",
    "function_code_fastcall",
    "instance_call",
    "instancemethod_call",
    "instancemethod_call",
    "methoddescr_call",
    "proxy_call",
    "slot_tp_call",
    "trace_call_function",
    "type_call",
    "weakref_call",
    "wrap_call",
    "wrapper_call",
    "wrapperdescr_call",
}


@functools.lru_cache(maxsize=1000)
def _is_cpython_internal_symbol(symbol: str, file: str) -> bool:
    if "PyEval_EvalFrameEx" in symbol or "_PyEval_EvalFrameDefault" in symbol:
        is_candidate = True
    elif symbol.startswith(("PyEval", "_Py")):
        is_candidate = True
    elif "vectorcall" in symbol.lower():
        is_candidate = True
    elif symbol in SYMBOL_IGNORELIST:
        is_candidate = True
    elif "Objects/call.c" in file:
        is_candidate = True
    else:
        is_candidate = False

    if is_candidate:
        return re.search(RE_CPYTHON_PATHS, file) is not None
    return False


def is_cpython_internal(frame: StackFrame) -> bool:
    symbol, file, _ = frame
    return _is_cpython_internal_symbol(symbol, file)


def is_frame_interesting(frame: StackFrame) -> bool:
    function, file, _ = frame

    if file.endswith("runpy.py") or file == "<frozen runpy>":
        return False

    return not _is_cpython_internal_symbol(function, file)


def is_frame_from_import_system(frame: StackFrame) -> bool:
    function, file, _ = frame
    if "frozen importlib" in file:
        return True
    if function in {"import_name", "import_from", "import_all_from"} and file.endswith(
        "ceval.c"
    ):
        return True
    return False
